from .exception_base import ExceptionBase


# --------------------------------------------------------------------------------
class CommunicationExceptionBase(ExceptionBase):
    def __init__(self, message: str):
        super().__init__(message)


# --------------------------------------------------------------------------------
class ClientDisconnectedException(CommunicationExceptionBase):
    def __init__(self, message: str = "Client disconnected"):
        super().__init__(message)


# --------------------------------------------------------------------------------
class IncompleteMessageException(CommunicationExceptionBase):
    def __init__(self, request_str: str):
        super().__init__(f'Incomplete request received: {request_str}. Closing connection.')


# --------------------------------------------------------------------------------
class CommunicationsRetryable(CommunicationExceptionBase):
    def __init__(self, message: str = "A retryable communications failure occurred."):
        super().__init__(message)


# --------------------------------------------------------------------------------
class CommunicationsNonRetryable(CommunicationExceptionBase):
    def __init__(self, message: str = "A non-retryable communications failure occurred."):
        super().__init__(message)


# --------------------------------------------------------------------------------
class CommunicationsMaxRetriesReached(CommunicationExceptionBase):
    def __init__(self, message: str = "Maximum retries exceeded in sending request."):
        super().__init__(message)


# --------------------------------------------------------------------------------
class CommunicationsEmptyResponse(CommunicationExceptionBase):
    def __init__(self, message: str = "Empty response received."):
        super().__init__(message)



from typing import Type, TypeVar
from pydantic import BaseModel as PydanticBaseModel

from .request_router import RequestRouter
from .queued_handler import QueuedHandler

from ..state_keepers.queued_handler_keeper import QueuedHandlerKeeper
from ..data_transfer_objects import EmptyDto

_T = TypeVar("_T", bound=PydanticBaseModel)


def queued_endpoint(
    route_key       : str,
    request_dto_type: Type[_T] = EmptyDto,
    page_size       : int = 100,
    max_retries     : int = 0,
):
    def inner_decorator(function):
        router                = RequestRouter()
        queued_handler_keeper = QueuedHandlerKeeper()

        new_handler = QueuedHandler[_T](
            route_key,
            function,
            request_dto_type,
            page_size,
            max_retries
        )
        router.register_handler(new_handler)
        queued_handler_keeper.add_queued_handler(new_handler)
        return function
    return inner_decorator



import asyncio
import copy
import logging
import numpy as np
import time

from typing import Any, Dict, List, Tuple

from ..util import SingletonType
from ..queues import PaginatedQueue
from ..configuration.config_models import AppConfiguration

# --------------------------------------------------------------------------------
class StatisticsKeeper(metaclass=SingletonType):
    __config              : AppConfiguration          = AppConfiguration()
    __running             : bool                      = False
    __logger              : logging.Logger            = logging.getLogger()
    __gather_period       : int                       = 300
    __history_length      : int                       = 12
    __start_time          : float                     = time.time()
    __statistics_current  : Dict[str, Any]            = {}
    __statistics_history  : List[Dict[str, Any]]      = []
    __persisted_queues    : Dict[str, PaginatedQueue] = {}
    __endpoint_durations  : Dict[str, List[float]]    = {}

    def __init__(self):
        pass

    def set_gather_period(self, gather_period: int):
        self.__gather_period = gather_period

    def set_history_length(self, history_length: int):
        self.__history_length = history_length

    async def get_endpoint_percentiles(self) -> Dict[str, Tuple[float, float]]:
        retval: Dict[str, Tuple[float, float]] = {}
        for key, duration_list in self.__endpoint_durations.items():
            if len(duration_list) > 0:
                p95 = float(np.percentile(duration_list, 95))
                p99 = float(np.percentile(duration_list, 99))
                retval[key] = (p95, p99)
            else:
                retval[key] = (-1.0, -1.0)
        return retval

    def __reset_endpoint_durations(self):
        for key in self.__endpoint_durations.keys():
            self.__endpoint_durations[key] = []

    async def __update_current_statistics(self):
        current_time = time.time()

        self.__statistics_current['timestamp']               = current_time
        self.__statistics_current['uptime']                  = current_time - self.__start_time
        self.__statistics_current['application']             = {}
        self.__statistics_current['application']['name']     = self.__config.name
        self.__statistics_current['application']['instance'] = self.__config.instance

        endpoint_percentiles = await self.get_endpoint_percentiles()
        for key, percentiles in endpoint_percentiles.items():
            self.set_statistic_value(f"endpoint_data.{key}.p95", float(percentiles[0]))
            self.set_statistic_value(f"endpoint_data.{key}.p99", float(percentiles[1]))

        for key in self.__persisted_queues.keys():
            self.set_statistic_value(key, self.__persisted_queues[key].size())

    async def get_current_statistics(self) -> Dict[str, Any]:
        await self.__update_current_statistics()
        return self.__statistics_current

    async def get_last_gathered_statistics(self) -> Dict[str, Any]:
        if len(self.__statistics_history) > 0:
            return self.__statistics_history[0]
        else:
            return {}

    async def get_full_gathered_statistics(self) -> List[Dict[str, Any]]:
        return self.__statistics_history

    def set_statistic_value(self, key: str, value: float):
        keys = key.split('.')
        self.__deep_set(self.__statistics_current, keys, value)

    @staticmethod
    def __deep_get(dictionary, keys):
        for key in keys:
            if isinstance(dictionary, dict):
                dictionary = dictionary.get(key, None)
            else:
                return None
        return dictionary

    @staticmethod
    def __deep_set(dictionary, keys, value):
        for key in keys[:-1]:
            dictionary = dictionary.setdefault(key, {})
        dictionary[keys[-1]] = value

    def __get_statistic_value(self, dictionary, keys) -> float:
        value = self.__deep_get(dictionary, keys)
        if value is None:
            return 0
        else:
            return value

    def increment(self, key: str, value: float = 1):
        keys          = key.split('.')
        current_value = self.__get_statistic_value(self.__statistics_current, keys)
        self.__deep_set(self.__statistics_current, keys, current_value+value)

    def decrement(self, key: str, value: float = 1):
        keys          = key.split('.')
        current_value = self.__get_statistic_value(self.__statistics_current, keys)
        self.__deep_set(self.__statistics_current, keys, current_value-value)

    def add_persisted_queue(self, key: str, queue: PaginatedQueue):
        self.__persisted_queues[key] = queue

    def track_endpoint_data(self, key: str):
        self.increment(f"endpoint_data.{key}.call_count", 0)
        self.__endpoint_durations[key] = []

    def add_endpoint_stats(self, key: str, percentile: float = 0):
        self.increment(f"endpoint_data.{key}.call_count")
        self.__endpoint_durations[key].append(percentile)

    async def __aenter__(self):
        self.__start_time = int(time.time())
        self.__running    = True
        self.__logger.info("Starting stats gathering.")

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.__running = False
        self.__logger.info("Stopping stats gathering.")

    def __reset_stats(self, dictionary):
        for key, value in dictionary.items():
            if isinstance(value, dict):
                self.__reset_stats(value)
            else:
                dictionary[key] = 0

    async def __reset_current_statistics(self):
        self.__reset_endpoint_durations()
        self.__reset_stats(self.__statistics_current)

    async def gather_statistics(self):
        while self.__running:
            sleep_count: int = 0

            while self.__running and sleep_count < self.__gather_period:
                await asyncio.sleep(1)
                sleep_count += 1

            if self.__running:
                self.__logger.info(f"Gathering statistics")
                await self.__update_current_statistics()
                self.__statistics_history.insert(0, copy.deepcopy(self.__statistics_current))
                if len(self.__statistics_history) > self.__history_length:
                    self.__statistics_history.pop()
                await self.__reset_current_statistics()



from .empty import EmptyDto

from .json_protocol import RequestDTO
from .json_protocol import ResponseDTO

from .queued_endpoint_response import QueuedEndpointResponseDTO
from .statistics import StatsRequestDto, StatsResponseDto



import argparse

from typing import List


# --------------------------------------------------------------------------------
class SmartFormatter(argparse.HelpFormatter):
    def _split_lines(self, text: str, width: int) -> List[str]:
        lines: List[str] = []
        for line_str in text.split('\n'):
            line: List[str] = []
            line_len = 0
            for word in line_str.split():
                word_len = len(word)
                next_len = line_len + word_len
                if line:
                    next_len += 1
                if next_len > width:
                    lines.append(' '.join(line))
                    line.clear()
                    line_len = 0
                elif line:
                    line_len += 1

                line.append(word)
                line_len += word_len

            lines.append(' '.join(line))
        return lines

    def _fill_text(self, text: str, width: int, indent: str) -> str:
        return '\n'.join(indent + line for line in self._split_lines(text, width - len(indent)))


argument_parser: argparse.ArgumentParser = argparse.ArgumentParser(
    formatter_class=SmartFormatter
)


# --------------------------------------------------------------------------------
argument_parser.add_argument(
    "-st", "--server_type",
    required = True,
    choices  = ["tcp", "udp", "uds"],
    type     = str,
    default  = None,
    help     = "The type of server you want to interact with."
)

argument_parser.add_argument(
    "-sd", "--server_details",
    metavar  = "<server details>",
    required = True,
    default  = None,
    help     = "Connection details for the server.\n"
               "- For TCP or UDP:\n"
               "- : This should be a string in the format [HOST:PORT].\n"
               "- For UDS:\n"
               "- : This should be a path to the socket file."
)

argument_parser.add_argument(
    "-ac", "--action",
    required = False,
    choices  = ["stat", "qem", "qsm"],
    type     = str,
    default  = "stat",
    help     = "The action you want to perform.\n"
               " - 'stat': get statistics\n"
               " - 'qem' : Queued Endpoint Management.\n"
               " - 'qes' : Queued Sender Management.\n"
               "(default = 'stat')"
)

# Stats or queue group
# --------------------------------------------------------------------------------
stat_or_queue_group = argument_parser.add_mutually_exclusive_group()

stat_or_queue_group.add_argument(
    "-stat", "--statistics_type",
    required = False,
    choices  = ["current", "gathered", "full"],
    type     = str,
    default  = "current",
    help     = "The statistics type you want to retrieve.\n"
               "- 'current': The current statistical period.\n"
               "- 'gathered': The last gathered statistical period.\n"
               "- 'full': The full gathered statistics history.\n"
               " (default = 'current')"
)

stat_or_queue_group.add_argument(
    "-rk" , "--route_key",
    metavar = "<route_key>",
    default = None,
    type    = str,
    help    = "The route key of the queue you wish to interact with."
)

# --------------------------------------------------------------------------------
queue_one_action_group = argument_parser.add_mutually_exclusive_group()

queue_one_action_group.add_argument("-dt" , "--data"              , required=False, action="store_true", help="Get data about a queue.")

# Queued Endpoints Only
queue_one_action_group.add_argument("-pr" , "--pause_receiving"   , required=False, action="store_true", help="For Queued Endpoints Only:\n- Pause receiving on a queue.")
queue_one_action_group.add_argument("-pp" , "--pause_processing"  , required=False, action="store_true", help="For Queued Endpoints Only:\n- Pause processing on a queue.")
queue_one_action_group.add_argument("-ur" , "--unpause_receiving" , required=False, action="store_true", help="For Queued Endpoints Only:\n- UN-Pause receiving on a queue.")
queue_one_action_group.add_argument("-up" , "--unpause_processing", required=False, action="store_true", help="For Queued Endpoints Only:\n- UN-Pause processing on a queue.")

queue_one_action_group.add_argument("-pa" , "--pause_all"         , required=False, action="store_true", help="For Queued Endpoints:\n- Pause both receiving and processing.\nFor Queued Senders:\n- Pause sending.")
queue_one_action_group.add_argument("-ua" , "--unpause_all"       , required=False, action="store_true", help="For Queued Endpoints:\n- UN-Pause both receiving and processing.\nFor Queued Senders:\n- UN-Pause sending.")
queue_one_action_group.add_argument("-e10", "--error_10"          , required=False, action="store_true", help="Get the first 10 uuids in an error database of a queue.")
queue_one_action_group.add_argument("-rp1", "--reprocess_one"     , required=False, action="store_true", help="Reprocess a specified entry in an error database of a queue.\n- [Requires -rid].")
queue_one_action_group.add_argument("-rpa", "--reprocess_all"     , required=False, action="store_true", help="Reprocess all entries in an error queue.")

# The danger zone
# --------------------------------------------------------------------------------
queue_one_action_group.add_argument("-ins", "--inspect_request"   , required=False, action="store_true", help="View a request on an error database of a queue.\n- [Requires -rid].")
queue_one_action_group.add_argument("-pop", "--pop_request"       , required=False, action="store_true", help="Pop a request from an error database of a queue.\n- [Requires -rid].")
queue_one_action_group.add_argument("-clr", "--clear"             , required=False, action="store_true", help="Clear the error database of a queue completely.\n- WARNING: All requests in the error queue are deleted!")

# Request id
# --------------------------------------------------------------------------------
argument_parser.add_argument(
    "-rid", "--request_uid",
    metavar = "<request uuid>",
    default = None,
    type    = str,
    help    = "The UUID for an item in a queue."
)

command_line_args: argparse.Namespace = argument_parser.parse_args()



import uuid

from typing import Dict, List, Any

from ..requests.queued_handler import QueuedHandler
from ..util import SingletonType


# --------------------------------------------------------------------------------
class QueuedHandlerKeeper(metaclass=SingletonType):
    __queued_handlers: Dict[str, QueuedHandler] = {}

    def __init__(self):
        pass

    # --------------------------------------------------------------------------------
    def has_route_key(self, route_key: str) -> bool:
        if route_key not in self.__queued_handlers.keys():
            return False
        return True

    # --------------------------------------------------------------------------------
    def add_queued_handler(self, queued_handler: QueuedHandler):
        self.__queued_handlers[queued_handler.get_route_key()] = queued_handler

    # --------------------------------------------------------------------------------
    async def get_queue_information(self, route_key: str) -> Dict[str, Any] | None:
        if route_key not in self.__queued_handlers.keys():
            return None
        return await self.__get_queue_information(route_key)

    # --------------------------------------------------------------------------------
    async def __get_queue_information(self, route_key: str) -> Dict[str, Any]:
        return {
            "pending"          : await self.__queued_handlers[route_key].pending_queue_size(),
            "error"            : await self.__queued_handlers[route_key].error_queue_size(),
            "receiving_paused" : self.__queued_handlers[route_key].is_receiving_paused(),
            "processing_paused": self.__queued_handlers[route_key].is_processing_paused()
        }

    # --------------------------------------------------------------------------------
    async def get_first_10_error_uuids(self, route_key) -> List[str] | None:
        if route_key not in self.__queued_handlers.keys():
            return None
        return await self.__queued_handlers[route_key].get_first_x_error_uuids(10)

    # --------------------------------------------------------------------------------
    async def pause_receiving_for_queued_handler(self, route_key: str):
        if route_key not in self.__queued_handlers.keys():
            return None
        self.__queued_handlers[route_key].pause_receiving()
        return await self.__get_queue_information(route_key)

    # --------------------------------------------------------------------------------
    async def pause_processing_for_queued_handler(self, route_key: str):
        if route_key not in self.__queued_handlers.keys():
            return None
        self.__queued_handlers[route_key].pause_processing()
        return await self.__get_queue_information(route_key)

    # --------------------------------------------------------------------------------
    async def pause_all_for_queued_handler(self, route_key: str):
        if route_key not in self.__queued_handlers.keys():
            return None
        self.__queued_handlers[route_key].pause_receiving()
        self.__queued_handlers[route_key].pause_processing()
        return await self.__get_queue_information(route_key)

    # --------------------------------------------------------------------------------
    async def unpause_receiving_for_queued_handler(self, route_key: str):
        if route_key not in self.__queued_handlers.keys():
            return None
        self.__queued_handlers[route_key].unpause_receiving()
        return await self.__get_queue_information(route_key)

    # --------------------------------------------------------------------------------
    async def unpause_processing_for_queued_handler(self, route_key: str):
        if route_key not in self.__queued_handlers.keys():
            return None
        self.__queued_handlers[route_key].unpause_processing()
        return await self.__get_queue_information(route_key)

    # --------------------------------------------------------------------------------
    async def unpause_all_for_queued_handler(self, route_key: str):
        if route_key not in self.__queued_handlers.keys():
            return None
        self.__queued_handlers[route_key].unpause_receiving()
        self.__queued_handlers[route_key].unpause_processing()
        return await self.__get_queue_information(route_key)

    # --------------------------------------------------------------------------------
    async def pop_request_from_error_queue(self, route_key: str, request_uid: uuid.UUID):
        if route_key not in self.__queued_handlers.keys():
            return None
        result = await self.__queued_handlers[route_key].pop_request_from_error_queue(request_uid)
        if not result:
            return False
        queue_information = await self.__get_queue_information(route_key)
        queue_information["request_data"] = result
        return queue_information

    # --------------------------------------------------------------------------------
    async def inspect_request_from_error_queue(self, route_key: str, request_uid: uuid.UUID):
        if route_key not in self.__queued_handlers.keys():
            return None
        result = await self.__queued_handlers[route_key].inspect_request_in_error_queue(request_uid)
        if not result:
            return False
        queue_information = await self.__get_queue_information(route_key)
        queue_information["request_data"] = result.dict()
        return queue_information

    # --------------------------------------------------------------------------------
    async def reprocess_error_queue(self, route_key: str):
        if route_key not in self.__queued_handlers.keys():
            return None
        await self.__queued_handlers[route_key].reprocess_error_queue()
        return await self.__get_queue_information(route_key)

    # --------------------------------------------------------------------------------
    async def clear_error_queue(self, route_key: str):
        if route_key not in self.__queued_handlers.keys():
            return None
        await self.__queued_handlers[route_key].error_queue_clear()
        return await self.__get_queue_information(route_key)

    # --------------------------------------------------------------------------------
    async def reprocess_error_queue_request_uid(self, route_key: str, request_uid: uuid.UUID):
        if route_key not in self.__queued_handlers.keys():
            return None
        result = await self.__queued_handlers[route_key].reprocess_error_queue_request_uid(request_uid)
        if not result:
            return False
        queue_information = await self.__get_queue_information(route_key)
        queue_information["request_data"] = result.dict()
        return queue_information



from typing import Type
from pydantic import BaseModel as PydanticBaseModel

from .request_router import RequestRouter
from .standard_handler import StandardHandler

from ..data_transfer_objects import EmptyDto


def endpoint(route_key: str, request_dto_type: Type[PydanticBaseModel] = EmptyDto):
    def inner_decorator(function):
        router      = RequestRouter()
        new_handler = StandardHandler(route_key, function, request_dto_type)
        router.register_handler(new_handler)
        return function
    return inner_decorator



import asyncio
import socket

from ..stream_server_base import StreamServerBase

from ...configuration.config_models import ConfigUDS


# --------------------------------------------------------------------------------
class UDSServer(StreamServerBase):
    def __init__(self, configuration : ConfigUDS):
        super().__init__()
        self.__server_path  : str  = f"{configuration.directory}/{configuration.socket_file_name}"
        self.__uds_supported: bool = hasattr(socket, "AF_UNIX")

    # --------------------------------------------------------------------------------
    async def __aenter__(self):
        if self.__uds_supported:
            try:
                self._running    = True
                self._logger.info(f"Setting up UDS server.")
                await self.__setup_server()
            except Exception as e:
                self._running = False
                self._server.close()
                raise e
        else:
            self._logger.info(f"UDS not supported on this platform, will not start server.")

    # --------------------------------------------------------------------------------
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.__uds_supported:
            self.stop()

    # --------------------------------------------------------------------------------
    def stop(self):
        if self._running:
            self._running = False
            self._logger.info(f"Stopping UDS server for {self.__server_path}.")
            self._server.close()

    # --------------------------------------------------------------------------------
    async def __setup_server(self):
        self._server   = await asyncio.start_unix_server(self._handle_request, self.__server_path)
        serving_address = self._server.sockets[0].getsockname()
        self._logger.info(f'Serving UDS on {serving_address}')

    # --------------------------------------------------------------------------------
    async def serve(self):
        if self._running:
            async with self._server:
                try:
                    await self._server.serve_forever(),
                except asyncio.exceptions.CancelledError:
                    self._logger.info("UDS Server cancelled.")


# --------------------------------------------------------------------------------
if __name__ == "__main__":
    print("Not an executable script, intended for use in other scripts")



import json
import logging
import timeit

from pydantic import ValidationError

from ..data_transfer_objects import RequestDTO, ResponseDTO
from ..requests.request_router import RequestRouter, RoutingExceptionBase
from ..requests.status import Status
from ..state_keepers.statistics_keeper import StatisticsKeeper

# --------------------------------------------------------------------------------
class ServerBase:
    def __init__(self):
        self._running          : bool             = False
        self._logger           : logging.Logger   = logging.getLogger()
        self._request_router   : RequestRouter    = RequestRouter()
        self._statistics_keeper: StatisticsKeeper = StatisticsKeeper()

    # --------------------------------------------------------------------------------
    async def _route_request(self, request_data: str) -> ResponseDTO:
        request_uuid: str = ""
        try:
            start_time   = timeit.default_timer()
            request_dict = json.loads(request_data.strip())
            request      = RequestDTO(**request_dict)
            request_uuid = request.uid
            response     = await self._request_router.route_request(request)
            end_time     = timeit.default_timer() - start_time
            self._statistics_keeper.add_endpoint_stats(request.route_key, end_time)
            return ResponseDTO(
                uid    = request_uuid,
                status = Status.SUCCESS.value,
                data   = response
            )
        except json.decoder.JSONDecodeError as e:
            return ResponseDTO(
                uid    = request_uuid,
                status = Status.PROTOCOL_PARSING_ERROR.value,
                data   = str(e)
            )
        except ValidationError as e:
            return ResponseDTO(
                uid    = request_uuid,
                status = Status.PYDANTIC_VALIDATION_ERROR.value,
                data   = str(e)
            )
        except RoutingExceptionBase as e:
            return ResponseDTO(
                uid    = request_uuid,
                status = e.status,
                data   = e.message
            )
        except Exception as e:
            return ResponseDTO(
                uid    = request_uuid,
                status = Status.UNKNOWN.value,
                data   = str(e)
            )

# --------------------------------------------------------------------------------
if __name__ == "__main__":
    print("Not an executable script, intended for use in other scripts")



from typing import Any, List
from pydantic import BaseModel as PydanticBaseModel


# --------------------------------------------------------------------------------
class QManagementRequestDto(PydanticBaseModel):
    queue_route_key: str


# --------------------------------------------------------------------------------
class QManagementItemRequestDto(QManagementRequestDto):
    request_uid    : str


# --------------------------------------------------------------------------------
class QDatabaseSizesDto(PydanticBaseModel):
    pending: int
    error  : int


# --------------------------------------------------------------------------------
class QueuedEndpointInformationDto(PydanticBaseModel):
    route_key        : str
    receiving_paused : bool
    processing_paused: bool
    database_sizes   : QDatabaseSizesDto


# --------------------------------------------------------------------------------
class QueuedSenderInformationDto(PydanticBaseModel):
    route_key           : str
    send_process_paused : bool
    database_sizes      : QDatabaseSizesDto


# --------------------------------------------------------------------------------
class QManagementResponseDto(PydanticBaseModel):
    queue_data  : QueuedEndpointInformationDto | QueuedSenderInformationDto | List[str] | None = None
    request_data: Any | None = None
    message     : str



import json

from ...data_transfer_objects import StatsRequestDto, StatsResponseDto
from ...clients.client_base import ClientBase
from ...sending.sender_base import SenderBase


# --------------------------------------------------------------------------------
def print_response(stat_type: str, response: StatsResponseDto):
    print("================================================================================")
    print(f"Retrieved statistics: [{stat_type}]")
    print("--------------------------------------------------------------------------------")
    print(f"{json.dumps(response.statistics, indent=4)}")
    print("================================================================================")


# --------------------------------------------------------------------------------
class EcoStatisticsGetSender(SenderBase[StatsRequestDto, StatsResponseDto]):
    def __init__(self, client: ClientBase):
        super().__init__(
            client,
            "eco.statistics.get",
            StatsRequestDto,
            StatsResponseDto
        )

    async def send(self, stat_type: str = "current"):
        request_data = StatsRequestDto(type=stat_type)
        response     = await self.send_data(request_data)
        print_response(stat_type, response)



from .communication import (
    IncompleteMessageException,
    ClientDisconnectedException,
    CommunicationsNonRetryable,
    CommunicationsMaxRetriesReached,
    CommunicationsEmptyResponse
)

from .response import (
    ResponseException,
    ProtocolParsingException,
    ClientDeniedException,
    PydanticValidationException,
    RouteKeyUnknownException,
    ServerBusyException,
    ProcessingException,
    UnknownException,
    UnknownStatusCodeException,
)



from .queued_management_sender_base import QueueManagementSenderBase, QueueManagementItemSenderBase

from ...clients import ClientBase


# --------------------------------------------------------------------------------
class EcoQueuedHandlerDataSender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_handler.data")


# --------------------------------------------------------------------------------
class EcoQueuedHandlerReceivingPauseSender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_handler.receiving.pause")


# --------------------------------------------------------------------------------
class EcoQueuedHandlerProcessingPauseSender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_handler.processing.pause")


# --------------------------------------------------------------------------------
class EcoQueuedHandlerAllPauseSender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_handler.all.pause")


# --------------------------------------------------------------------------------
class EcoQueuedHandlerReceivingUnPauseSender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_handler.receiving.unpause")


# --------------------------------------------------------------------------------
class EcoQueuedHandlerProcessingUnPauseSender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_handler.processing.unpause")


# --------------------------------------------------------------------------------
class EcoQueuedHandlerAllUnPauseSender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_handler.all.unpause")


# --------------------------------------------------------------------------------
class EcoQueuedHandlerErrorsGetFirst10Sender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_handler.errors.get_first_10")


# --------------------------------------------------------------------------------
class EcoQueuedHandlerErrorsReprocessAllSender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_handler.errors.reprocess.all")


# --------------------------------------------------------------------------------
class EcoQueuedHandlerErrorsClearSender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_handler.errors.clear")


# --------------------------------------------------------------------------------
class EcoQueuedHandlerErrorsReprocessOneSender(QueueManagementItemSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_handler.errors.reprocess.one")


# --------------------------------------------------------------------------------
class EcoQueuedHandlerErrorsPopRequestSender(QueueManagementItemSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_handler.errors.pop_request")


# --------------------------------------------------------------------------------
class EcoQueuedHandlerErrorsInspectRequestSender(QueueManagementItemSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_handler.errors.inspect_request")



import uuid

from typing import Any, Dict, cast
from pydantic import BaseModel as PydanticBaseModel

from ..requests.endpoint import endpoint
from ..state_keepers.queued_handler_keeper import QueuedHandlerKeeper
from ..util.utility_functions import string_to_uuid
from ..data_transfer_objects.queue_management import (
    QManagementRequestDto,
    QManagementItemRequestDto,
    QDatabaseSizesDto,
    QueuedEndpointInformationDto,
    QManagementResponseDto
)

# --------------------------------------------------------------------------------
def make_queued_endpoint_information_dto(route_key: str, queue_info_dict: Dict[str, Any]) -> QueuedEndpointInformationDto:
    return QueuedEndpointInformationDto(
        route_key         = route_key,
        receiving_paused  = queue_info_dict["receiving_paused"],
        processing_paused = queue_info_dict["processing_paused"],
        database_sizes    = QDatabaseSizesDto(
            pending = queue_info_dict["pending"],
            error   = queue_info_dict["error"]
        )
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_handler.data", QManagementRequestDto)
async def eco_queued_handler_data(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                  = cast(QManagementRequestDto, request)
    queued_handler_keeper = QueuedHandlerKeeper()
    queue_info_dict       = await queued_handler_keeper.get_queue_information(data.queue_route_key)
    if not queue_info_dict:
        return QManagementResponseDto(data=None, message=f"No queue for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data   = make_queued_endpoint_information_dto(data.queue_route_key, queue_info_dict),
        request_data = None,
        message      = f"Queued Endpoint[{data.queue_route_key}]: Data retrieved.",
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_handler.receiving.pause", QManagementRequestDto)
async def eco_queued_handler_receiving_pause(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                  = cast(QManagementRequestDto, request)
    queued_handler_keeper = QueuedHandlerKeeper()
    queue_info_dict       = await queued_handler_keeper.pause_receiving_for_queued_handler(data.queue_route_key)
    if not queue_info_dict:
        return QManagementResponseDto(message=f"No queue for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data = make_queued_endpoint_information_dto(data.queue_route_key, queue_info_dict),
        message    = f"Queued Endpoint[{data.queue_route_key}]: Paused RECEIVING.",
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_handler.receiving.unpause", QManagementRequestDto)
async def eco_queued_handler_receiving_unpause(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                  = cast(QManagementRequestDto, request)
    queued_handler_keeper = QueuedHandlerKeeper()
    queue_info_dict       = await queued_handler_keeper.unpause_receiving_for_queued_handler(data.queue_route_key)
    if not queue_info_dict:
        return QManagementResponseDto(message=f"No queue for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data = make_queued_endpoint_information_dto(data.queue_route_key, queue_info_dict),
        message    = f"Queued Endpoint[{data.queue_route_key}]: UN-Paused RECEIVING.",
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_handler.processing.pause", QManagementRequestDto)
async def eco_queued_handler_processing_pause(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                  = cast(QManagementRequestDto, request)
    queued_handler_keeper = QueuedHandlerKeeper()
    queue_info_dict       = await queued_handler_keeper.pause_processing_for_queued_handler(data.queue_route_key)
    if not queue_info_dict:
        return QManagementResponseDto(message=f"No queue for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data = make_queued_endpoint_information_dto(data.queue_route_key, queue_info_dict),
        message    = f"Queued Endpoint[{data.queue_route_key}]: Paused PROCESSING.",
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_handler.processing.unpause", QManagementRequestDto)
async def eco_queued_handler_processing_unpause(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                  = cast(QManagementRequestDto, request)
    queued_handler_keeper = QueuedHandlerKeeper()
    queue_info_dict       = await queued_handler_keeper.unpause_processing_for_queued_handler(data.queue_route_key)
    if not queue_info_dict:
        return QManagementResponseDto(message=f"No queue for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data = make_queued_endpoint_information_dto(data.queue_route_key, queue_info_dict),
        message    = f"Queued Endpoint[{data.queue_route_key}]: UN-Paused PROCESSING.",
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_handler.all.pause", QManagementRequestDto)
async def eco_queued_handler_all_pause(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                  = cast(QManagementRequestDto, request)
    queued_handler_keeper = QueuedHandlerKeeper()
    queue_info_dict       = await queued_handler_keeper.pause_all_for_queued_handler(data.queue_route_key)
    if not queue_info_dict:
        return QManagementResponseDto(message=f"No queue for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data = make_queued_endpoint_information_dto(data.queue_route_key, queue_info_dict),
        message    = f"Queued Endpoint[{data.queue_route_key}]: Paused ALL.",
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_handler.all.unpause", QManagementRequestDto)
async def eco_queued_handler_all_unpause(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                  = cast(QManagementRequestDto, request)
    queued_handler_keeper = QueuedHandlerKeeper()
    queue_info_dict       = await queued_handler_keeper.unpause_all_for_queued_handler(data.queue_route_key)
    if not queue_info_dict:
        return QManagementResponseDto(message=f"No queue for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data = make_queued_endpoint_information_dto(data.queue_route_key, queue_info_dict),
        message    = f"Queued Endpoint[{data.queue_route_key}]: UN-Paused ALL.",
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_handler.errors.get_first_10", QManagementRequestDto)
async def eco_queued_handler_errors_get_first_10(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                  = cast(QManagementRequestDto, request)
    queued_handler_keeper = QueuedHandlerKeeper()
    uuid_list             = await queued_handler_keeper.get_first_10_error_uuids(data.queue_route_key)
    if uuid_list is None:
        return QManagementResponseDto(message=f"No queue for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data = uuid_list,
        message    = f"Queued Endpoint[{data.queue_route_key}]: Retrieved first 10 UUIDs for error database."
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_handler.errors.reprocess.all", QManagementRequestDto)
async def eco_queued_handler_errors_reprocess_all(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                  = cast(QManagementRequestDto, request)
    queued_handler_keeper = QueuedHandlerKeeper()
    queue_info_dict       = await queued_handler_keeper.reprocess_error_queue(data.queue_route_key)
    if not queue_info_dict:
        return QManagementResponseDto(message=f"No queue for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data = make_queued_endpoint_information_dto(data.queue_route_key, queue_info_dict),
        message    = f"Queued Endpoint[{data.queue_route_key}]: Error database entries, moved to incoming database."
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_handler.errors.clear", QManagementRequestDto)
async def eco_queued_handler_errors_clear(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                  = cast(QManagementRequestDto, request)
    queued_handler_keeper = QueuedHandlerKeeper()
    queue_info_dict       = await queued_handler_keeper.clear_error_queue(data.queue_route_key)
    if not queue_info_dict:
        return QManagementResponseDto(message=f"No queue for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data = make_queued_endpoint_information_dto(data.queue_route_key, queue_info_dict),
        message    = f"Queued Endpoint[{data.queue_route_key}]: Error database cleared."
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_handler.errors.reprocess.one", QManagementItemRequestDto)
async def eco_queued_handler_errors_reprocess_one(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                  = cast(QManagementItemRequestDto, request)
    queued_handler_keeper = QueuedHandlerKeeper()
    request_uid           = string_to_uuid(data.request_uid)
    if not request_uid:
        return QManagementResponseDto(message=f"[{data.request_uid}] is not a valid UUID.")

    queue_info_dict = await queued_handler_keeper.reprocess_error_queue_request_uid(data.queue_route_key, uuid.UUID(data.request_uid))

    if queue_info_dict is None:
        return QManagementResponseDto(message=f"No queue for route key: [{data.queue_route_key}]")
    elif not queue_info_dict:
        return QManagementResponseDto(message=f"No request with uid [{data.request_uid}] in error queue for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data   = make_queued_endpoint_information_dto(data.queue_route_key, queue_info_dict),
        request_data = queue_info_dict["request_data"],
        message      = f"Queued Endpoint[{data.queue_route_key}]: Error database entry[{data.request_uid}] moved to incoming database."
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_handler.errors.pop_request", QManagementItemRequestDto)
async def eco_queued_handler_errors_pop_request(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                  = cast(QManagementItemRequestDto, request)
    queued_handler_keeper = QueuedHandlerKeeper()
    request_uid           = string_to_uuid(data.request_uid)
    if not request_uid:
        return QManagementResponseDto(message=f"[{data.request_uid}] is not a valid UUID.")

    queue_info_dict = await queued_handler_keeper.pop_request_from_error_queue(data.queue_route_key, uuid.UUID(data.request_uid))
    if queue_info_dict is None:
        return QManagementResponseDto(message=f"No queue for route key: [{data.queue_route_key}]")
    elif not queue_info_dict:
        return QManagementResponseDto(message=f"No request with uid [{data.request_uid}] in error queue for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data   = make_queued_endpoint_information_dto(data.queue_route_key, queue_info_dict),
        request_data = queue_info_dict["request_data"],
        message      = f"Queued Endpoint[{data.queue_route_key}]: POPPED Error database entry[{data.request_uid}]."
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_handler.errors.inspect_request", QManagementItemRequestDto)
async def eco_queued_handler_errors_inspect_request(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                  = cast(QManagementItemRequestDto, request)
    queued_handler_keeper = QueuedHandlerKeeper()
    request_uid           = string_to_uuid(data.request_uid)
    if not request_uid:
        return QManagementResponseDto(message=f"[{data.request_uid}] is not a valid UUID.")

    queue_info_dict = await queued_handler_keeper.inspect_request_from_error_queue(data.queue_route_key, uuid.UUID(data.request_uid))
    if queue_info_dict is None:
        return QManagementResponseDto(message=f"No queue for route key: [{data.queue_route_key}]")
    elif not queue_info_dict:
        return QManagementResponseDto(message=f"No request with uid [{data.request_uid}] in error queue for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data   = make_queued_endpoint_information_dto(data.queue_route_key, queue_info_dict),
        request_data = queue_info_dict["request_data"],
        message      = f"Queued Endpoint[{data.queue_route_key}]: INSPECTING Error database entry[{data.request_uid}]."
    )



import os
import sys
import json
import platform

from typing import Dict, Any
from pydantic import BaseModel as PydanticBaseModel, Field

from .argument_parser import command_line_args

from ..util.singleton import SingletonType
from ..util.utility_functions import camel_to_snake
from ..exceptions.exception_base import ExceptionBase

# --------------------------------------------------------------------------------
class ConfigurationExceptionBase(ExceptionBase):
    def __init__(self, message: str):
        super().__init__(message)

# --------------------------------------------------------------------------------
class InstanceConfigurationNotFoundException(ConfigurationExceptionBase):
    def __init__(self, name: str, instance: str):
        super().__init__(f"Configuration for instance [{instance}] of application [{name}] not found!")

# --------------------------------------------------------------------------------
application_name     = camel_to_snake(os.path.basename(sys.argv[0]).replace('.py', ''))
application_instance = command_line_args.instance
env_prefix           = "ECOENV"
env_global           = f"{env_prefix}"
env_app_name         = f"{application_name.upper()}"
env_instance         = application_instance.upper()
env_app_instance     = f"{env_app_name}_{env_instance}"

# --------------------------------------------------------------------------------
# Used for when we want to check instance, application and global default setting
# --------------------------------------------------------------------------------
def get_eco_env(postfix: str, default: Any = None, instance_level_only: bool = False):
    global env_app_instance, env_app_name, env_prefix
    global_env_name    = f"{env_prefix}_{postfix}"
    app_level_env_name = f"{global_env_name}_{env_app_name}"
    instance_env_name  = f"{app_level_env_name}_{env_instance}"
    retval: str = os.environ.get(f"{instance_env_name}") # First we try to get the value for the running instance
    if not retval:
        if instance_level_only:
            return default
        retval = os.environ.get(f"{app_level_env_name}") # Now we try to get an application level default
        if not retval:
            retval = os.environ.get(f"{global_env_name}") # And here we try to get a global default
            if not retval:
                return default # If we still don't have a value, then we return the specified default
    return retval

# --------------------------------------------------------------------------------
# Used for when only an instance level setting would be valid
# --------------------------------------------------------------------------------
def get_instance_env(postfix: str, default: Any = None):
    return get_eco_env(postfix, default, True)

# --------------------------------------------------------------------------------
def get_platform_default_dir():
    if platform.system() == "Windows":
        return get_eco_env("DEFAULT_DIR", "C:\\")
    else:
        return get_eco_env("DEFAULT_DIR", "/tmp")

platform_default_dir = get_platform_default_dir()

# ConfigStatisticsKeeper
# --------------------------------------------------------------------------------
def get_statistics_gather_period():
    # 300 seconds, i.e. 5 minutes by default
    return int(get_eco_env("STAT_GP", 300))

def get_statistics_history_length():
    # 12 by default
    # i.e. Keep an hours worth of statistics if the gather period is 300 seconds
    return int(get_eco_env("STAT_HL", 12))

class ConfigStatisticsKeeper(PydanticBaseModel):
    gather_period : int = Field(default_factory=get_statistics_gather_period)
    history_length: int = Field(default_factory=get_statistics_history_length)

# ConfigLoggingFile
# --------------------------------------------------------------------------------
def get_logfile_directory():
    return get_eco_env("LOG_DIR", platform_default_dir)

def get_logfile_base_file_name():
    return f"{application_name}-{application_instance}"

def get_logfile_base_file_path():
    return f"{get_logfile_directory()}/{get_logfile_base_file_name()}.log"

def get_buffer_size():
    # Default is no buffer. i.e. 0
    return int(get_eco_env("LOG_BUF_SIZE", 0))

def get_logfile_max_size_in_bytes():
    # (1024*1024)*10 = 10,485,760 i.e. 10 mega-bytes
    return int(get_eco_env("LOG_MAX_SIZE", 10485760))

def get_logfile_max_files():
    # 10 Files by default
    return int(get_eco_env("LOG_MAX_FILES", 10))

class ConfigLoggingFile(PydanticBaseModel):
    directory         : str  = Field(default_factory=get_logfile_directory)
    base_file_name    : str  = Field(default_factory=get_logfile_base_file_name)
    base_file_path    : str  = Field(default_factory=get_logfile_base_file_path)
    buffer_size       : int  = Field(default_factory=get_buffer_size)
    max_size_in_bytes : int  = Field(default_factory=get_logfile_max_size_in_bytes)
    max_files         : int  = Field(default_factory=get_logfile_max_files)

# ConfigLogging
# --------------------------------------------------------------------------------
def get_logging_date_format():
    return get_eco_env("LOG_DATE_FORMAT", '%Y%m%d%H%M%S') # Default to '%Y%m%d%H%M%S'

def get_logging_format():
    return get_eco_env("LOG_FORMAT", '%(asctime)s.%(msecs)03d|%(levelname)s|%(filename)s|%(lineno)d|%(message)s')

def get_logging_level():
    # List of options here: 'debug', 'info', 'warn', 'error', 'critical'
    return get_eco_env("LOG_LEVEL", 'info')

class ConfigLogging(PydanticBaseModel):
    # By default, we want both console and file logging.
    # This is because it's too easy for users to forget, that starting an
    # application in the back-ground, means the terminal can be closed and all
    # logging for the application is lost.
    # Therefore, we force the user to make a choice, and remain conscious of
    # the fact that they are making a choice.
    console_only: bool              = command_line_args.log_console_only
    file_only   : bool              = command_line_args.log_file_only
    file_logging: ConfigLoggingFile = ConfigLoggingFile()
    date_format : str               = Field(default_factory=get_logging_date_format)
    format      : str               = Field(default_factory=get_logging_format)
    level       : str               = Field(default_factory=get_logging_level)

# ConfigTCP
# --------------------------------------------------------------------------------
class ConfigTCP(PydanticBaseModel):
    host: str = "127.0.0.1"
    port: int = 8888

# ConfigUDP
# --------------------------------------------------------------------------------
class ConfigUDP(PydanticBaseModel):
    host: str = "127.0.0.1"
    port: int = 8889

# ConfigUDS
# --------------------------------------------------------------------------------
class ConfigUDS(PydanticBaseModel):
    directory       : str = "/tmp" # because we don't want sock files surviving reboot
    socket_file_name: str = "DEFAULT"

# ConfigApplicationInstance
# --------------------------------------------------------------------------------
def get_app_instance_lock_dir():
    return get_eco_env("LOCK_DIR", platform_default_dir)

def parse_host_port_string(value: str):
    # TODO: Throw an exception if we can't parse the string!
    host, port = value.split(":")
    if not host:
        return None
    if not port:
        return None
    return {
        "host": host,
        "port": int(port)
    }

def get_app_instance_tcp():
    tcp_string = get_instance_env("TCP", None)
    if tcp_string:
        host_port_data = parse_host_port_string(tcp_string)
        if host_port_data:
            return ConfigTCP(host=host_port_data["host"], port=host_port_data["port"])
    return None

def get_app_instance_udp():
    udp_string = get_instance_env("UDP", None)
    if udp_string:
        host_port_data = parse_host_port_string(udp_string)
        if host_port_data:
            return ConfigUDP(host=host_port_data["host"], port=host_port_data["port"])
    return None

def get_app_instance_uds():
    uds_string = get_instance_env("UDS", None)
    if not uds_string:
        return None
    directory        = os.path.dirname(uds_string)
    socket_file_name = os.path.basename(uds_string)
    if not directory or not socket_file_name:
        return None
    if not os.path.isdir(directory):
        return None
    if socket_file_name == "DEFAULT":
        socket_file_name = f"{application_name}_{application_instance}.uds.sock"
    return ConfigUDS(directory=directory, socket_file_name=socket_file_name)

def get_app_instance_queue_directory():
    # Do not default queue_directory to anything. Make the user explicitly set it.
    # Rather let them get an error, than lose their queues due to them having been
    # written to a place that gets cleaned on reboot.
    return get_eco_env("QUEUE_DIR", None)

def get_app_instance_extra():
    extra_env_prefix = f"{env_prefix}_EXTRA_{env_app_instance}_"
    extra_config     = {
        k: v for k, v in os.environ.items()
        if k.startswith(extra_env_prefix)
    }
    retval: Dict[str, Any] = {}
    for key in extra_config.keys():
        new_key = key.replace(extra_env_prefix, "")
        retval[new_key] = extra_config[key]
    return retval

class ConfigApplicationInstance(PydanticBaseModel):
    application_name: str                    = application_name
    instance_id     : str                    = application_instance
    stats_keeper    : ConfigStatisticsKeeper = ConfigStatisticsKeeper()
    logging         : ConfigLogging          = ConfigLogging()
    lock_directory  : str                    = Field(default_factory=get_app_instance_lock_dir)
    tcp             : ConfigTCP              = Field(default_factory=get_app_instance_tcp)
    udp             : ConfigUDP              = Field(default_factory=get_app_instance_udp)
    uds             : ConfigUDS              = Field(default_factory=get_app_instance_uds)
    queue_directory : str                    = Field(default_factory=get_app_instance_queue_directory)
    extra           : Any                    = Field(default_factory=get_app_instance_extra)

# ConfigApplicationInstanceDefaults
# --------------------------------------------------------------------------------
class ConfigApplicationInstanceDefaults(PydanticBaseModel):
    lock_directory  : str                    = Field(default_factory=get_app_instance_lock_dir)
    logging         : ConfigLogging          = ConfigLogging()
    stats_keeper    : ConfigStatisticsKeeper = ConfigStatisticsKeeper()
    queue_directory : str                    = Field(default_factory=get_app_instance_queue_directory)
    extra           : Any                    = None

# ConfigApplication
# --------------------------------------------------------------------------------
class ConfigApplication(PydanticBaseModel):
    name             : str                                  = application_name
    running_instance : str                                  = application_instance
    instances        : Dict[str, ConfigApplicationInstance] = {application_instance: ConfigApplicationInstance()}
    instance_defaults: ConfigApplicationInstanceDefaults    = ConfigApplicationInstanceDefaults()

# If a configuration file was specified, load the config from there.
# --------------------------------------------------------------------------------
try:
    if command_line_args.config_file:
        file_path = command_line_args.config_file
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"The specified configuration file [{file_path}] does NOT exist.")
        if not os.path.isfile(file_path):
            raise FileNotFoundError(f"The specified configuration file [{file_path}] is NOT a file.")

        with open(file_path, 'r') as f:
            config_json = json.load(f)

        application_configuration = ConfigApplication(**config_json)
    else:
        application_configuration = ConfigApplication()
except Exception as e:
    print(str(e))
    sys.exit(1)

if application_instance not in application_configuration.instances.keys():
    # When loading from environment, this isn't even be possible.
    # It's here because loading configurations from file, could have this happen.
    print(f"Configuration for instance [{application_instance}] of application [{application_name}] not found!")
    sys.exit(2)

instance_configuration = application_configuration.instances[application_instance]

# --------------------------------------------------------------------------------
class AppConfiguration(metaclass=SingletonType):
    name             = application_name
    instance         = application_instance
    tcp              = instance_configuration.tcp
    udp              = instance_configuration.udp
    uds              = instance_configuration.uds
    stats_keeper     = instance_configuration.stats_keeper
    logging          = instance_configuration.logging
    lock_directory   = instance_configuration.lock_directory
    queue_directory  = instance_configuration.queue_directory
    extra            = instance_configuration.extra

    def dict(self):
        return {
            "name"           : self.name,
            "instance"       : self.instance,
            "tcp"            : None if self.tcp is None else self.tcp.model_dump(),
            "udp"            : None if self.udp is None else self.udp.model_dump(),
            "uds"            : None if self.uds is None else self.uds.model_dump(),
            "stats_keeper"   : self.stats_keeper.model_dump(),
            "logging"        : self.logging.model_dump(),
            "lock_directory" : self.lock_directory,
            "queue_directory": self.queue_directory,
            "extra"          : self.extra,
        }



import asyncio

from abc import ABC, abstractmethod
from typing import Tuple

from .client_base import ClientBase

from ..exceptions import (
    CommunicationsNonRetryable,
    CommunicationsMaxRetriesReached,
    CommunicationsEmptyResponse,
)

# TODO: Check if connection is still open
# --------------------------------------------------------------------------------
class PersistentStreamClientBase(ClientBase, ABC):
    def __init__(
        self,
        max_retries: int   = 3,
        retry_delay: float = 0.1,
    ):
        super().__init__(max_retries, retry_delay)
        self.__connected: bool                 = False
        self.__reader   : asyncio.StreamReader = None
        self.__writer   : asyncio.StreamWriter = None
        self.__read_lock : asyncio.Lock   = asyncio.Lock()
        self.__write_lock: asyncio.Lock   = asyncio.Lock()

    # --------------------------------------------------------------------------------
    @abstractmethod
    async def open_connection(self) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        pass

    # --------------------------------------------------------------------------------
    async def close_connection(self):
        self.__writer.close()
        await self.__writer.wait_closed()

    # --------------------------------------------------------------------------------
    async def _send_message(self, request: str) -> str:
        if not self.__connected:
            self.__reader, self.__writer = await self.open_connection()
            self.__connected = True

        async with self.__write_lock:
            self.__writer.write(request.encode())

        async with self.__read_lock:
            data = await self.__reader.readline()

        if not data:
            raise CommunicationsEmptyResponse()

        response_str = data.decode()
        return response_str

    # --------------------------------------------------------------------------------
    async def _send_message_retry_loop(self, request: str) -> str:
        while self.retry_count < self.max_retries and not self.success:
            try:
                response_str = await self._send_message(request)
                self.success = True
                return response_str
            except (TimeoutError, ConnectionResetError, BrokenPipeError, asyncio.TimeoutError) as e:
                self.retry_count += 1
                if self.retry_count >= self.max_retries:
                    raise CommunicationsMaxRetriesReached()
                else:
                    await asyncio.sleep(self.retry_delay)
            except Exception as e:
                raise CommunicationsNonRetryable(str(e))



import uuid

from pydantic import BaseModel as PydanticBaseModel
from abc import ABC, abstractmethod
from typing import Type

from ..data_transfer_objects import EmptyDto


# --------------------------------------------------------------------------------
class HandlerBase(ABC):
    def __init__(self, route_key: str, request_dto_type: Type[PydanticBaseModel] = EmptyDto):
        self._route_key      : str                     = route_key
        self.request_dto_type: Type[PydanticBaseModel] = request_dto_type

    def get_route_key(self) -> str:
        return self._route_key

    async def attempt_request(self, request_uuid: uuid.UUID, request_data) -> PydanticBaseModel:
        request_data = self.request_dto_type(**request_data)
        return await self.run(request_uuid, request_data)

    @abstractmethod
    async def run(self, request_uuid: uuid.UUID, request_data) -> PydanticBaseModel:
        pass


# --------------------------------------------------------------------------------
if __name__ == "__main__":
    print("Not an executable script, intended for use in other scripts")



import argparse

argument_parser: argparse.ArgumentParser = argparse.ArgumentParser()

argument_parser.add_argument(
    "-i", "--instance",
    type     = str,
    required = True,
    help     = 'The instance id this invocation of the script needs to run as.',
)

argument_parser.add_argument(
    "-c", "--config_file",
    type     = str,
    required = False,
    default  = None,
    help     = "Specify a configuration file to load for this invocation of the application. Default is: None",
)

logging_group = argument_parser.add_mutually_exclusive_group()

logging_group.add_argument(
    "-lco", "--log_console_only",
    required = False,
    action   = "store_true",
    help     = "Set logging too console only. i.e. Don't log to a file.",
)

logging_group.add_argument(
    "-lfo", "--log_file_only",
    required = False,
    action   = "store_true",
    help     = "Set logging too file only. i.e. Don't log to console.",
)

command_line_args = argument_parser.parse_args()



import uuid
from typing import Type
from pydantic import BaseModel as PydanticBaseModel
from .handler_base import HandlerBase


class StandardHandler(HandlerBase):
    def __init__(
        self,
        route_key: str,
        function,
        request_dto_type: Type[PydanticBaseModel]
    ):
        super().__init__(route_key, request_dto_type)
        self.function = function

    async def run(self, request_uuid: uuid.UUID, request_data) -> PydanticBaseModel:
        return await self.function(request_uuid, request_data)



import asyncio

from .server_base import ServerBase

# --------------------------------------------------------------------------------
class StreamServerBase(ServerBase):
    def __init__(self):
        super().__init__()
        self._server       : asyncio.Server = None
        self.__ENQ_byte    : int            =  5 # Decimal  5 = Ascii ENQ (enquiry) character
        self.__ACK_byte    : int            =  6 # Decimal  6 = Ascii ACK (acknowledge) character
        self.__LF_byte     : int            = 10 # Decimal 10 = Ascii LF (line feed) character = '\n'
        self.__ACK_response: bytes          = bytes([self.__ACK_byte, self.__LF_byte])
        self.__read_lock   : asyncio.Lock   = asyncio.Lock()
        self.__write_lock  : asyncio.Lock   = asyncio.Lock()

    # --------------------------------------------------------------------------------
    async def __read_data(self, reader: asyncio.StreamReader):
        async with self.__read_lock:
            return await reader.readline()

    # --------------------------------------------------------------------------------
    async def __write_data(self, writer: asyncio.StreamWriter, data: bytes):
        async with self.__write_lock:
            writer.write(data)
            await writer.drain()

    # TODO: Check client against white-list!
    # --------------------------------------------------------------------------------
    async def _handle_request(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        if not self._running:
            return
        while True: # We keep the connection open.
            try:
                bytes_read = await reader.readline()

                # Check if the client closed the connection.
                # Take note: An incomplete read due to EOF is treated as a client disconnect,
                # So we check if the last byte read is '\n' i.e. Decimal 10/Ascii symbol: LF
                if not bytes_read or bytes_read[-1] != self.__LF_byte:
                    break

                if bytes_read[0] == self.__ENQ_byte: # The client is asking if we are still connected.
                    await self.__write_data(writer, self.__ACK_response)
                else:
                    response_dict = await self._route_request(bytes_read.decode())
                    await self.__write_data(
                        writer,
                        (response_dict.model_dump_json() + '\n').encode()
                    )
            except ConnectionResetError:
                self._logger.info("Connection reset by peer")
                break
        writer.close()



from enum import Enum


# --------------------------------------------------------------------------------
class Status(Enum):
    # All is good
    SUCCESS                   = 0   # The request was successfully processed.
    # Client is broken
    PROTOCOL_PARSING_ERROR    = 100 # A message received was so broken we could not do basic JSON parsing on it.
    CLIENT_DENIED             = 200 # The Client is not in the white list and my not communicate to this process.
    PYDANTIC_VALIDATION_ERROR = 300 # Pydantic could not validate a DTO, from the JSON we received.
    ROUTE_KEY_UNKNOWN         = 400 # No matching identifier for the supplied value of "route_key" in request, was found
    # Server is broken
    APPLICATION_BUSY          = 500 # used by server applications to inform requesters that a request won't be processed due to server overload.
    PROCESSING_FAILURE        = 600 # For use by apps that need to report on an internal processing failure.
    UNKNOWN                   = 999 # Sorry, I don't know what went wrong.


# --------------------------------------------------------------------------------
if __name__ == "__main__":
    print("Not an executable script, intended for use in other scripts")



import uuid

from typing import Type, TypeVar, Generic
from pydantic import BaseModel as PydanticBaseModel
from .queued_handler_base import QueuedRequestHandlerBase

_T = TypeVar("_T", bound=PydanticBaseModel)


class QueuedHandler(Generic[_T], QueuedRequestHandlerBase[_T]):
    def __init__(
        self,
        route_key       : str,
        function,
        request_dto_type: Type[_T],
        page_size       : int = 100,
        max_retries     : int = 0,
    ):
        super().__init__(
            route_key,
            request_dto_type,
            page_size,
            max_retries
        )
        self.function = function

    # --------------------------------------------------------------------------------
    async def process_queued_request(self,  request_uuid: uuid.UUID, request: _T) -> bool:
        return await self.function(request_uuid, request)



from pydantic import BaseModel as PydanticBaseModel


# --------------------------------------------------------------------------------
class QueuedEndpointResponseDTO(PydanticBaseModel):
    uid: str



import asyncio
import uuid
import logging

from typing import Type, TypeVar, Generic, List
from pydantic import BaseModel as PydanticBaseModel

from .sender_base import SenderBase

from ..clients import ClientBase
from ..data_transfer_objects import EmptyDto
from ..queues.pending_queue import PendingQueue
from ..state_keepers.statistics_keeper import StatisticsKeeper
from ..exceptions import ServerBusyException, CommunicationsMaxRetriesReached

_RequestDTOType  = TypeVar("_RequestDTOType" , bound=PydanticBaseModel)
_ResponseDTOType = TypeVar("_ResponseDTOType", bound=PydanticBaseModel)


# --------------------------------------------------------------------------------
class QueuedSenderBase(Generic[_RequestDTOType, _ResponseDTOType], SenderBase[_RequestDTOType, _ResponseDTOType]):
    def __init__(
        self,
        client           : ClientBase,
        route_key        : str,
        request_dto_type : Type[_RequestDTOType],
        response_dto_type: Type[_ResponseDTOType] = EmptyDto,
        wait_period      : float                  = 0,
        page_size        : int                    = 100,
        max_retries      : int                    = 0,
    ):
        super().__init__(
            client,
            route_key,
            request_dto_type,
            response_dto_type
        )

        self.running                  : bool             = False
        self._sending_paused          : bool             = True
        self.statistics_keeper        : StatisticsKeeper = StatisticsKeeper()
        self.log                      : logging.Logger   = logging.getLogger()
        self.__send_process_scheduled : bool             = False
        self.wait_period              : float            = wait_period
        self.page_size                : int              = page_size
        self.max_retries              : int              = max_retries
        self.shutdown                 : bool             = False
        self.queue                    : PendingQueue     = None
        self.on_shutdown_future       : asyncio.Future   = None
        self.__send_process_task      : asyncio.Task     = None

    # --------------------------------------------------------------------------------
    async def send(self, *args, **kwargs):
        pass

    # --------------------------------------------------------------------------------
    async def enqueue(self, request_data: _RequestDTOType, request_uuid: uuid.UUID = None) -> None:
        if not request_uuid:
            uuid_to_use = uuid.uuid4()
        else:
            uuid_to_use = request_uuid
        await self.queue.push_pending(uuid_to_use, request_data, 0)
        self.__check_process_send_queue()

    # --------------------------------------------------------------------------------
    def __check_process_send_queue(self):
        if self.__send_process_task is None:
            self.__send_process_task = asyncio.create_task(self.__process_send_queue())
        elif self.__send_process_task.done():
            self.__send_process_task = asyncio.create_task(self.__process_send_queue())

    # --------------------------------------------------------------------------------
    def shut_down(self):
        self.pause_send_process()
        self.shutdown = True
        self.__shut_down_check()

    # --------------------------------------------------------------------------------
    def __shut_down_check(self):
        if not self.running and self.shutdown:
            route_key = self.get_route_key()
            self.log.info(f"Queued sender [{route_key}] stopping.")
            self.queue.shut_down()
            self.on_shutdown_future.set_result(True)
            self.log.info(f"Queued sender [{route_key}] stopped.")

    # --------------------------------------------------------------------------------
    async def wait_for_shutdown(self):
        while not self.on_shutdown_future.done():
            await asyncio.sleep(1)

    # --------------------------------------------------------------------------------
    async def __do_queue_processing(self):
        while not self._sending_paused and self.queue.pending_q.size():
            queued_item  = await self.queue.pop()
            request_uid  = uuid.UUID(queued_item.uid)
            request_data = self._request_dto_type(**queued_item.data)
            retries      = queued_item.retries
            try:
                if self.wait_period > 0:
                    await asyncio.sleep(self.wait_period)
                await self.send_data(request_data, request_uid)
            except (ServerBusyException, CommunicationsMaxRetriesReached): # Only retry sending, if the sending is retryable
                retries += 1
                if retries >= self.max_retries:
                    await self.queue.push_error(request_uid, request_data, "Max retries reached.")
                else:
                    await self.queue.push_pending(request_uid, request_data, retries)
            except Exception as e : # Move it to the error queue
                await self.queue.push_error(request_uid, request_data, str(e))

    # --------------------------------------------------------------------------------
    async def __process_send_queue(self) -> None:
        try:
            self.running = True
            await self.__do_queue_processing()
        finally:
            self.running                  = False
            self.__send_process_scheduled = False
            self.__shut_down_check()

    # --------------------------------------------------------------------------------
    def pause_send_process(self):
        self._sending_paused = True

    def unpause_send_process(self):
        self._sending_paused = False
        self.__check_process_send_queue()

    def is_send_process_paused(self):
        return self._sending_paused

    # --------------------------------------------------------------------------------
    async def pending_queue_size(self) -> int:
        return await self.queue.get_pending_size()

    # --------------------------------------------------------------------------------
    async def error_queue_size(self) -> int:
        return await self.queue.get_error_size()

    # --------------------------------------------------------------------------------
    async def error_queue_clear(self):
        await self.queue.clear_error_queue()

    # --------------------------------------------------------------------------------
    async def get_first_x_error_uuids(self, how_many: int = 1) -> List[str]:
        return await self.queue.get_first_x_error_uuids(how_many)

    # --------------------------------------------------------------------------------
    async def pop_request_from_error_queue(self, request_uid: uuid.UUID):
        queued_request = await self.queue.pop_error_q_uuid(request_uid)
        if not queued_request:
            return None
        return self._request_dto_type(**queued_request)

    # --------------------------------------------------------------------------------
    async def inspect_request_in_error_queue(self, request_uid: uuid.UUID):
        queued_request = await self.queue.inspect_error_q_uuid(request_uid)
        if not queued_request:
            return None
        return self._request_dto_type(**queued_request)

    # --------------------------------------------------------------------------------
    async def reprocess_error_queue(self):
        self._sending_paused = True
        await self.queue.move_all_error_to_pending()
        self.__check_process_send_queue()
        self._sending_paused = False

    # --------------------------------------------------------------------------------
    async def reprocess_error_queue_request_uid(self, request_uid: uuid.UUID) -> _RequestDTOType | None:
        self._sending_paused = True
        queued_request = await self.queue.move_one_error_to_pending(request_uid)
        self.__check_process_send_queue()
        self._sending_paused = False
        return self._request_dto_type(**queued_request)

    # --------------------------------------------------------------------------------
    def __configure_queue(
        self,
        directory       : str,
        application_name: str,
        instance_id     : str,
    ):
        app_instance_string = f"{application_name}-{instance_id}"
        self.queue = PendingQueue(
            directory,
            f"{app_instance_string}-{self._route_key}-sender",
            self.page_size
        )
        self.statistics_keeper.add_persisted_queue(f"queued_sender_sizes.{self._route_key}.pending", self.queue.pending_q)
        self.statistics_keeper.add_persisted_queue(f"queued_sender_sizes.{self._route_key}.error"  , self.queue.error_q)

    # --------------------------------------------------------------------------------
    async def setup(
        self,
        directory: str,
        application_name: str,
        instance_id     : str,
    ):
        route_key = self.get_route_key()
        self.log.info(f"Queued sender [{route_key}] setup.")
        self.__configure_queue(directory, application_name, instance_id)
        self.unpause_send_process()

        self.__check_process_send_queue()

        loop = asyncio.get_running_loop()
        self.on_shutdown_future = loop.create_future()
        self.log.info(f"Queued sender [{route_key}] setup complete.")



import uuid

from abc import ABC, abstractmethod
from typing import Type, TypeVar, Generic
from pydantic import BaseModel as PydanticBaseModel

from ..clients import ClientBase
from ..data_transfer_objects import EmptyDto

_RequestDTOType  = TypeVar("_RequestDTOType" , bound=PydanticBaseModel)
_ResponseDTOType = TypeVar("_ResponseDTOType", bound=PydanticBaseModel)


# --------------------------------------------------------------------------------
class SenderBase(Generic[_RequestDTOType, _ResponseDTOType], ABC):
    def __init__(
        self,
        client           : ClientBase,
        route_key        : str,
        request_dto_type : Type[_RequestDTOType],
        response_dto_type: Type[_ResponseDTOType] = EmptyDto,
    ):
        self._client           : ClientBase             = client
        self._route_key        : str                    = route_key
        self._request_dto_type : Type[_RequestDTOType]  = request_dto_type
        self._response_dto_type: Type[_ResponseDTOType] = response_dto_type

    def get_route_key(self) -> str:
        return self._route_key

    @abstractmethod
    async def send(self, *args, **kwargs):
        pass

    async def send_data(self, data: _RequestDTOType, request_uid: uuid.UUID = None) -> _ResponseDTOType:
        return await self._client.send_message(
            self._route_key,
            data,
            self._response_dto_type,
            request_uid
        )



import uuid
from typing import Any, cast
from pydantic import BaseModel as PydanticBaseModel

from ..requests.endpoint import endpoint
from ..state_keepers.statistics_keeper import StatisticsKeeper
from ..data_transfer_objects import StatsRequestDto, StatsResponseDto

# --------------------------------------------------------------------------------
@endpoint("eco.statistics.get", StatsRequestDto)
async def eco_statistics_get(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data         = cast(StatsRequestDto, request)
    stats_keeper = StatisticsKeeper()

    if data.type == "current":
        stats = await stats_keeper.get_current_statistics()
    elif data.type == "full":
        stats = await stats_keeper.get_full_gathered_statistics()
    else:
        stats = await stats_keeper.get_last_gathered_statistics()

    return StatsResponseDto(statistics=stats)



import asyncio

from typing import Tuple
from ..persistent_stream_client_base import PersistentStreamClientBase

# --------------------------------------------------------------------------------
class PersistedTCPClient(PersistentStreamClientBase):
    def __init__(
        self,
        server_host: str,
        server_port: int,
        max_retries: int   = 3,
        retry_delay: float = 0.1,
    ):
        self.server_host: str   = server_host
        self.server_port: int   = server_port
        super().__init__(max_retries, retry_delay)

    # --------------------------------------------------------------------------------
    async def open_connection(self) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        return await asyncio.open_connection(self.server_host, self.server_port)



from .queued_management_sender_base import QueueManagementSenderBase, QueueManagementItemSenderBase

from ...clients import ClientBase


# --------------------------------------------------------------------------------
class EcoQueuedSenderDataSender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_sender.data")


# --------------------------------------------------------------------------------
class EcoQueuedSenderSendProcessPauseSender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_sender.send_process.pause")


# --------------------------------------------------------------------------------
class EcoQueuedSenderSendProcessUnPauseSender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_sender.send_process.unpause")


# --------------------------------------------------------------------------------
class EcoQueuedSenderErrorsGetFirst10Sender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_sender.errors.get_first_10")


# --------------------------------------------------------------------------------
class EcoQueuedSenderErrorsReprocessAllSender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_sender.errors.reprocess.all")


# --------------------------------------------------------------------------------
class EcoQueuedSenderErrorsClearSender(QueueManagementSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_sender.errors.clear")


# --------------------------------------------------------------------------------
class EcoQueuedSenderErrorsReprocessOneSender(QueueManagementItemSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_sender.errors.reprocess.one")


# --------------------------------------------------------------------------------
class EcoQueuedSenderErrorsPopRequestSender(QueueManagementItemSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_sender.errors.pop_request")


# --------------------------------------------------------------------------------
class EcoQueuedSenderErrorsInspectRequestSender(QueueManagementItemSenderBase):
    def __init__(self, client: ClientBase):
        super().__init__(client, "eco.queued_sender.errors.inspect_request")



import os
import asyncio
import signal
import argparse
import logging

from .logs import EcoLogger
from .configuration.config_models import AppConfiguration

from .requests.request_router import RequestRouter

from .servers import (
    TCPServer,
    UDPServer,
    UDSServer,
)

from .state_keepers.queued_sender_keeper import QueuedSenderKeeper
from .state_keepers.error_state_list import ErrorStateList
from .state_keepers.statistics_keeper import StatisticsKeeper

from .util import SingletonType

from .exceptions.exception_base import ExceptionBase

# Pycharm complains that we aren't using these imports,
# but the import is what does the work of getting everything up and
# running. So we do a noqa comment here.
from .standard_endpoints.statistics import eco_statistics_get # noqa
from .standard_endpoints.errors import eco_error_states_get, eco_error_states_clear # noqa
from .standard_endpoints.queued_handler_manager import ( # noqa
    eco_queued_handler_data,
    eco_queued_handler_errors_clear,
    eco_queued_handler_errors_get_first_10,
    eco_queued_handler_errors_inspect_request,
    eco_queued_handler_errors_pop_request,
    eco_queued_handler_errors_reprocess_all,
    eco_queued_handler_errors_reprocess_one,
    eco_queued_handler_processing_pause,
    eco_queued_handler_processing_unpause,
    eco_queued_handler_receiving_pause,
    eco_queued_handler_receiving_unpause,
)
from .standard_endpoints.queued_sender_manager import ( # noqa
    eco_queued_sender_data,
    eco_queued_sender_errors_clear,
    eco_queued_sender_errors_get_first_10,
    eco_queued_sender_errors_inspect_request,
    eco_queued_sender_errors_pop_request,
    eco_queued_sender_errors_reprocess_all,
    eco_queued_sender_errors_reprocess_one,
    eco_queued_sender_send_process_pause,
    eco_queued_sender_send_process_unpause,
)

# --------------------------------------------------------------------------------
class InstanceAlreadyRunningException(ExceptionBase):
    def __init__(
        self,
        application_name: str,
        instance_id     : str,
        process_id      : int
    ):
        super().__init__(f"Instance [{instance_id}] of [{application_name}] already running with process id [{process_id}]!")

# --------------------------------------------------------------------------------
class TerminationSignalException(Exception):
    pass

# --------------------------------------------------------------------------------
class ApplicationBase(metaclass=SingletonType):
    command_line_args     : argparse.Namespace      = None
    logger                : logging.Logger          = logging.getLogger()
    __eco_logger          : EcoLogger               = EcoLogger()
    __request_router      : RequestRouter           = RequestRouter()
    __statistics_keeper   : StatisticsKeeper        = StatisticsKeeper()
    __error_state_list    : ErrorStateList          = ErrorStateList()
    __running             : bool                    = False
    _configuration        : AppConfiguration        = AppConfiguration()
    __server_tcp          : TCPServer               = None
    __server_udp          : UDPServer               = None
    __server_uds          : UDSServer               = None

    # --------------------------------------------------------------------------------
    def __init__(self):
        self.__configure_basics()

    # --------------------------------------------------------------------------------
    @staticmethod
    def __handle_exit_signal(signum, frame):
        raise TerminationSignalException("Exit signal received")

    # --------------------------------------------------------------------------------
    def __setup_signal_handlers(self):
        for x_signal in (signal.SIGTERM, signal.SIGINT, signal.SIGHUP):
            signal.signal(x_signal, self.__handle_exit_signal)

    # --------------------------------------------------------------------------------
    def __enter__(self):
        self.__setup_signal_handlers()
        self.__running = True
        return self

    # --------------------------------------------------------------------------------
    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            if not isinstance(exc_val, asyncio.CancelledError) \
                    and not isinstance(exc_val, TerminationSignalException):
                import traceback
                self.logger.info(f"Shutdown: EXCEPTION: {exc_type}")
                traceback.print_exception(exc_type, exc_val, exc_tb)
            else:
                self.logger.info(f"Shutdown: Termination signal received.")
        else:
            self.logger.info(f"Shutdown: Application context ending.")

        self.__do_shutdown()
        return True

    # --------------------------------------------------------------------------------
    def __do_shutdown(self):
        self.logger.info("Doing Shutdown.")
        self.__stop_servers()
        self.__shut_down_queued_handlers()
        self.__shut_down_queued_senders()
        self.logger.info(f"Instance [{self._configuration.instance}] of application [{self._configuration.name}] shutdown.")
        self.__eco_logger.flush()

    # --------------------------------------------------------------------------------
    @staticmethod
    def __process_id_check(process_id: int) -> bool:
        try:
            os.kill(process_id, 0)
        except OSError:
            return False
        return True

    # --------------------------------------------------------------------------------
    @staticmethod
    def __create_lock_file(lock_file_path: str):
        process_id = os.getpid()
        with open(lock_file_path, "w") as lock_file:
            lock_file.write(f"{process_id}\n")

    # --------------------------------------------------------------------------------
    def __lock_file_check(self):
        lock_file_name  = f"{self._configuration.name}-{self._configuration.instance}.lock"
        lock_file_path  = f"{self._configuration.lock_directory}/{lock_file_name}"
        if os.path.exists(lock_file_path):
            with open(lock_file_path, "r") as lock_file:
                process_id = int(lock_file.readline())
            if self.__process_id_check(process_id):
                raise InstanceAlreadyRunningException(self._configuration.name, self._configuration.instance, process_id)
            os.remove(lock_file_path)
        self.__create_lock_file(lock_file_path)

    # --------------------------------------------------------------------------------
    def __configure_basics(self) -> None:
        self.__lock_file_check()

        self.__statistics_keeper.set_gather_period(
            self._configuration.stats_keeper.gather_period
        )
        self.__statistics_keeper.set_history_length(
            self._configuration.stats_keeper.history_length
        )

        self.__configure_communication_servers()

    # --------------------------------------------------------------------------------
    def __configure_communication_servers(self):
        if self._configuration.tcp:
            self.__server_tcp = TCPServer(self._configuration.tcp)

        if self._configuration.udp:
            self.__server_udp = UDPServer(self._configuration.udp)

        if self._configuration.uds:
            uds_config = self._configuration.uds
            if uds_config.socket_file_name == "DEFAULT":
                uds_config.socket_file_name = f"{self._configuration.name}_{self._configuration.instance}_uds.sock"
            self.__server_uds = UDSServer(uds_config)

    # --------------------------------------------------------------------------------
    async def __setup_queued_handlers(self):
        queue_directory = self._configuration.queue_directory
        for queued_handler in self.__request_router.get_queued_handlers():
            await queued_handler.setup(
                queue_directory,
                self._configuration.name,
                self._configuration.instance
            )

    # --------------------------------------------------------------------------------
    async def __setup_queued_senders(self):
        queued_sender_keeper = QueuedSenderKeeper()
        queue_directory = self._configuration.queue_directory
        for queued_senders in queued_sender_keeper.get_queued_senders():
            await queued_senders.setup(
                queue_directory,
                self._configuration.name,
                self._configuration.instance
            )

    # --------------------------------------------------------------------------------
    def __shut_down_queued_handlers(self):
        for queued_handler in self.__request_router.get_queued_handlers():
            queued_handler.shut_down()

    # --------------------------------------------------------------------------------
    @staticmethod
    def __shut_down_queued_senders():
        queued_sender_keeper = QueuedSenderKeeper()
        for queued_sender in queued_sender_keeper.get_queued_senders():
            queued_sender.shut_down()

    # --------------------------------------------------------------------------------
    def start(self):
        if not self.__running:
            self.logger.info("Application not running in context. Shutting down.")
            self.__eco_logger.flush()
            return
        asyncio.run(self.__start())

    # --------------------------------------------------------------------------------
    def stop(self):
        self.__running = False

    # --------------------------------------------------------------------------------
    async def __start_stats_keeper(self):
        async with self.__statistics_keeper:
            await self.__statistics_keeper.gather_statistics()

    # --------------------------------------------------------------------------------
    async def __start_tcp_server(self):
        if self.__server_tcp:
            async with self.__server_tcp:
                await self.__server_tcp.serve()

    # --------------------------------------------------------------------------------
    def __stop_tcp_server(self):
        if self.__server_tcp:
            self.__server_tcp.stop()

    # --------------------------------------------------------------------------------
    async def __start_udp_server(self):
        if self.__server_udp:
            async with self.__server_udp:
                await self.__server_udp.serve()

    # --------------------------------------------------------------------------------
    def __stop_udp_server(self):
        if self.__server_udp:
            self.__server_udp.stop()

    # --------------------------------------------------------------------------------
    async def __start_uds_server(self):
        if self.__server_uds:
            async with self.__server_uds:
                await self.__server_uds.serve()

    # --------------------------------------------------------------------------------
    def __stop_uds_server(self):
        if self.__server_uds:
            self.__server_uds.stop()

    # --------------------------------------------------------------------------------
    def __stop_servers(self):
        self.__stop_tcp_server()
        self.__stop_udp_server()
        self.__stop_uds_server()

    # --------------------------------------------------------------------------------
    async def __start(self):
        tasks = []

        await self.__setup_queued_handlers()
        await self.__setup_queued_senders()

        for queued_handler in self.__request_router.get_queued_handlers():
            task = asyncio.create_task(queued_handler.wait_for_shutdown())
            tasks.append(task)

        queued_sender_keeper = QueuedSenderKeeper()
        for queued_sender in queued_sender_keeper.get_queued_senders():
            task = asyncio.create_task(queued_sender.wait_for_shutdown())
            tasks.append(task)

        tasks.append(asyncio.create_task(self.__start_stats_keeper()))
        tasks.append(asyncio.create_task(self.__start_tcp_server()))
        tasks.append(asyncio.create_task(self.__start_udp_server()))
        tasks.append(asyncio.create_task(self.__start_uds_server()))

        self.__eco_logger.flush()
        await asyncio.gather(*tasks)

# --------------------------------------------------------------------------------
if __name__ == "__main__":
    print("Not an executable script, intended for use in other scripts")



import asyncio
import uuid
import logging

from pydantic import BaseModel as PydanticBaseModel
from abc import ABC, abstractmethod
from typing import Type, TypeVar, Generic, List

from .handler_base import HandlerBase
from .status import Status

from ..util.fire_and_forget_tasks import fire_and_forget_task
from ..data_transfer_objects import QueuedEndpointResponseDTO
from ..queues.pending_queue import PendingQueue
from ..state_keepers.statistics_keeper import StatisticsKeeper

_T = TypeVar("_T", bound=PydanticBaseModel)

# --------------------------------------------------------------------------------
class QueuedRequestHandlerExceptionBase(Exception):
    def __init__(self, status: int, message: str):
        self.status: int = status
        super().__init__(f"status: [{status}] message: [{message}]")

# --------------------------------------------------------------------------------
class QueuedRequestHandlerReceivingPausedException(QueuedRequestHandlerExceptionBase):
    def __init__(self, route_key: str):
        super().__init__(Status.APPLICATION_BUSY.value, f"Receiving on '{route_key}' has been paused.")

# --------------------------------------------------------------------------------
class QueuedRequestHandlerBase(Generic[_T], HandlerBase, ABC):
    def __init__(
        self,
        route_key       : str,
        request_dto_type: Type[_T],
        page_size       : int = 0,
        max_retries     : int = 0,
    ):
        super().__init__(route_key, request_dto_type)
        self.running                : bool             = False
        self._receiving_paused      : bool             = True
        self._processing_paused     : bool             = True
        self.__processing_scheduled : bool             = False
        self.statistics_keeper      : StatisticsKeeper = StatisticsKeeper()
        self.log                    : logging.Logger   = logging.getLogger()
        self.page_size              : int              = page_size
        self.max_retries            : int              = max_retries
        self.shutdown               : bool             = False
        self.queue                  : PendingQueue     = None
        self.on_shutdown_future     : asyncio.Future   = None

    # --------------------------------------------------------------------------------
    def pause_receiving(self):
        self._receiving_paused = True

    def unpause_receiving(self):
        self._receiving_paused = False
        self.__check_process_queue()

    def pause_processing(self):
        self._processing_paused = True

    def unpause_processing(self):
        self._processing_paused = False
        self.__check_process_queue()

    def is_receiving_paused(self):
        return self._receiving_paused

    def is_processing_paused(self):
        return self._processing_paused

    # --------------------------------------------------------------------------------
    async def pending_queue_size(self) -> int:
        return await self.queue.get_pending_size()

    # --------------------------------------------------------------------------------
    async def error_queue_size(self) -> int:
        return await self.queue.get_error_size()

    # --------------------------------------------------------------------------------
    async def error_queue_clear(self):
        await self.queue.clear_error_queue()

    # --------------------------------------------------------------------------------
    async def get_first_x_error_uuids(self, how_many: int = 1) -> List[str]:
        return await self.queue.get_first_x_error_uuids(how_many)

    # --------------------------------------------------------------------------------
    async def pop_request_from_error_queue(self, request_uid: uuid.UUID) -> _T | None:
        queued_request = await self.queue.pop_error_q_uuid(request_uid)
        if not queued_request:
            return None
        return self.request_dto_type(**queued_request)

    # --------------------------------------------------------------------------------
    async def inspect_request_in_error_queue(self, request_uid: uuid.UUID) -> _T | None:
        queued_request = await self.queue.inspect_error_q_uuid(request_uid)
        if not queued_request:
            return None
        return self.request_dto_type(**queued_request)

    # --------------------------------------------------------------------------------
    async def reprocess_error_queue(self):
        self._processing_paused = True
        await self.queue.move_all_error_to_pending()
        self.__check_process_queue()
        self._processing_paused = False

    # --------------------------------------------------------------------------------
    async def reprocess_error_queue_request_uid(self, request_uid: uuid.UUID) -> _T | None:
        self._processing_paused = True
        queued_request = await self.queue.move_one_error_to_pending(request_uid)
        self.__check_process_queue()
        self._processing_paused = False
        return self.request_dto_type(**queued_request)

    # --------------------------------------------------------------------------------
    def __configure_queue(
        self,
        directory       : str,
        application_name: str,
        instance_id     : str,
    ):
        app_instance_string = f"{application_name}-{instance_id}"
        self.queue = PendingQueue(
            directory,
            f"{app_instance_string}-{self._route_key}-endpoint",
            self.page_size
        )
        self.statistics_keeper.add_persisted_queue(f"queued_endpoint_sizes.{self._route_key}.pending", self.queue.pending_q)
        self.statistics_keeper.add_persisted_queue(f"queued_endpoint_sizes.{self._route_key}.error"  , self.queue.error_q)

    # --------------------------------------------------------------------------------
    async def setup(
        self,
        directory       : str,
        application_name: str,
        instance_id     : str,
    ):
        route_key = self.get_route_key()
        self.log.info(f"Queued handler [{route_key}] setup.")
        self.__configure_queue(directory, application_name, instance_id)
        self.unpause_receiving()
        self.unpause_processing()

        self.__check_process_queue()

        loop = asyncio.get_running_loop()
        self.on_shutdown_future = loop.create_future()
        self.log.info(f"Queued handler [{route_key}] setup complete.")

    # --------------------------------------------------------------------------------
    def shut_down(self):
        self.pause_receiving()
        self.pause_processing()
        self.shutdown = True
        self.__shut_down_check()

    # --------------------------------------------------------------------------------
    def __shut_down_check(self):
        if not self.running and self.shutdown:
            route_key = self.get_route_key()
            self.log.info(f"Queued handler [{route_key}] stopping.")
            self.queue.shut_down()
            self.on_shutdown_future.set_result(True)
            self.log.info(f"Queued handler [{route_key}] stopped.")

    # --------------------------------------------------------------------------------
    async def wait_for_shutdown(self):
        while not self.on_shutdown_future.done():
            await asyncio.sleep(1)

    # --------------------------------------------------------------------------------
    async def _process_queue(self):
        self.running = True
        while not self._processing_paused and await self.queue.has_pending():
            queued_item  = await self.queue.pop()
            request_uid  = uuid.UUID(queued_item.uid)
            request_data = self.request_dto_type(**queued_item.data)
            retries      = queued_item.retries
            if not await self.process_queued_request(request_uid, request_data):
                retries += 1
                if retries >= self.max_retries:
                    await self.queue.push_error(request_uid, request_data, "Max retries reached.")
                else:
                    await self.queue.push_pending(request_uid, request_data, retries)
        self.running                = False
        self.__processing_scheduled = False
        self.__shut_down_check()

    # --------------------------------------------------------------------------------
    @abstractmethod
    async def process_queued_request(self, request_uuid: uuid.UUID, request: _T) -> bool:
        pass

    # TODO: Investigate if we need process scheduled check here, as for queued sender?
    # --------------------------------------------------------------------------------
    def __check_process_queue(self):
        if not self.running and not self.__processing_scheduled:
            self.__processing_scheduled = True
            fire_and_forget_task(self._process_queue())

    # --------------------------------------------------------------------------------
    async def run(self, request_uuid: uuid.UUID, request_data) -> PydanticBaseModel:
        if self._receiving_paused:
            raise QueuedRequestHandlerReceivingPausedException(self._route_key)
        fire_and_forget_task(self.queue.push_pending(request_uuid, request_data, 0))
        response = QueuedEndpointResponseDTO(uid = str(request_uuid))
        self.__check_process_queue()
        return response



from logging.handlers import RotatingFileHandler, BufferingHandler

# --------------------------------------------------------------------------------
class BufferedRotatingFileHandler(BufferingHandler):
    def __init__(
        self,
        filename,
        mode        = 'a',
        buffer_size = 0,
        max_bytes   = 0,
        backup_count= 0,
        encoding    = None,
        delay       = False,
    ):
        super().__init__(buffer_size)
        self.rotating_file_handler = RotatingFileHandler(
            filename,
            mode,
            maxBytes    = max_bytes,
            backupCount = backup_count,
            encoding    = encoding,
            delay       = delay,
        )
        self.buffer      = []
        self.buffer_size = buffer_size

    def setFormatter(self, formatter):
        self.rotating_file_handler.setFormatter(formatter)

    def flush(self):
        for record in self.buffer:
            self.rotating_file_handler.emit(record)
        self.buffer = []

    def close(self):
        self.rotating_file_handler.close()
        super().close()



import uuid
import logging

from typing import Any, List
from pydantic import BaseModel as PydanticBaseModel
from .paginated_queue import PaginatedQueue

log = logging.getLogger()

# --------------------------------------------------------------------------------
class PendingEntry(PydanticBaseModel):
    uid    : str
    retries: int = 0
    data   : Any

# --------------------------------------------------------------------------------
class ErrorEntry(PydanticBaseModel):
    uid   : str
    data  : Any
    reason: str

# --------------------------------------------------------------------------------
class PendingQueue:
    pending_q: PaginatedQueue[PendingEntry] = None
    error_q  : PaginatedQueue[ErrorEntry]   = None

    def __init__(
        self,
        directory     : str,
        file_basename : str,
        page_size     : int = 100
    ) -> None:
        self.__setup_pending_q(f"{directory}/{file_basename}-pending.sqlite", page_size)
        self.__setup_error_q  (f"{directory}/{file_basename}-error.sqlite"  , page_size)

    # --------------------------------------------------------------------------------
    def shut_down(self):
        self.pending_q.shut_down()
        self.error_q.shut_down()

    # --------------------------------------------------------------------------------
    async def has_pending(self):
        return not self.pending_q.is_empty()

    async def get_pending_size(self):
        return self.pending_q.size()

    async def get_error_size(self):
        return self.error_q.size()

    async def get_sizes(self):
        return {
            "pending": await self.get_pending_size(),
            "error"  : await self.get_error_size()
        }

    # --------------------------------------------------------------------------------
    async def move_all_error_to_pending(self):
        while self.error_q.size() > 0:
            popped_error_entry = await self.error_q.pop()
            await self.push_pending(popped_error_entry.uid, popped_error_entry.data, 0)

    async def move_one_error_to_pending(self, item_uid: uuid.UUID):
        popped_error_entry = await self.pop_error_q_uuid(item_uid)
        if popped_error_entry:
            await self.push_pending(popped_error_entry.uid, popped_error_entry.data, 0)
            return popped_error_entry.data
        return None

    async def clear_error_queue(self):
        await self.error_q.clear()

    async def get_first_x_error_uuids(self, how_many: int = 1) -> List[str]:
        return await self.error_q.get_first_x_uuids(how_many)

    # --------------------------------------------------------------------------------
    @staticmethod
    async def _pop_using_uuid(queue: PaginatedQueue, item_uid: uuid.UUID):
        queued_request = await queue.pop_uuid(item_uid)
        if not queued_request:
            return None
        return queued_request.data

    async def pop_error_q_uuid(self, request_uid: uuid.UUID) -> Any | None:
        return await self._pop_using_uuid(self.error_q, request_uid)

    async def pop_pending_q_uuid(self, request_uid: uuid.UUID) -> Any | None:
        return await self._pop_using_uuid(self.pending_q, request_uid)

    # --------------------------------------------------------------------------------
    @staticmethod
    async def _inspect_using_uuid(queue: PaginatedQueue, item_uid: uuid.UUID):
        queued_request = await queue.inspect_uuid(item_uid)
        if not queued_request:
            return None
        return queued_request.data

    async def inspect_error_q_uuid(self, request_uid: uuid.UUID) -> Any | None:
        return await self._inspect_using_uuid(self.error_q, request_uid)

    async def inspect_pending_q_uuid(self, request_uid: uuid.UUID) -> Any | None:
        return await self._inspect_using_uuid(self.pending_q, request_uid)

    # --------------------------------------------------------------------------------
    async def push_pending(self, item_uuid: uuid.UUID, item_data, retries: int = 0):
        entry_to_queue = PendingEntry(
            uid     = str(item_uuid),
            data    = item_data,
            retries = retries,
        )
        await self.pending_q.push(entry_to_queue, item_uuid)

    # --------------------------------------------------------------------------------
    async def push_error(self, item_uuid: uuid.UUID, item_data, reason: str):
        log.warning(f"Pushing message to error queue [{item_uuid}] {reason}]")
        entry_to_queue = ErrorEntry(
            uid    = str(item_uuid),
            data   = item_data,
            reason = reason,
        )
        await self.error_q.push(entry_to_queue, item_uuid)

    # --------------------------------------------------------------------------------
    async def pop(self):
        return await self.pending_q.pop()

    # --------------------------------------------------------------------------------
    def __setup_pending_q(self, file_path: str, page_size: int):
        self.pending_q = PaginatedQueue[PendingEntry](file_path, PendingEntry, page_size)

    # --------------------------------------------------------------------------------
    def __setup_error_q(self, file_path: str, page_size: int):
        self.error_q = PaginatedQueue[ErrorEntry](file_path, ErrorEntry, page_size)



import json
from typing import Dict, List, Any

from .error_state_keeper import ErrorStateKeeper

from ..util.singleton import SingletonType


# --------------------------------------------------------------------------------
class ErrorStateList(metaclass=SingletonType):
    error_state_map: Dict[str, ErrorStateKeeper] = {}

    def __init__(self):
        pass

    def __str__(self) -> str:
        return self.to_json()

    def to_json(self) -> str:
        error_list   : List[ErrorStateKeeper]     = self.get_error_states()
        response_list: List[Dict[str, Any]] = []
        for error in error_list:
            response_list.append(error.to_dict())
        return json.dumps(response_list)

    def increment(self, error_id: str, description: str):
        error_id = error_id.upper()
        if error_id not in self.error_state_map.keys():
            self.error_state_map[error_id] = ErrorStateKeeper(error_id, description)

        self.error_state_map[error_id].increment()

    def clear_some(self, error_id: str, how_many: int):
        if error_id in self.error_state_map.keys():
            self.error_state_map[error_id].clear_some(how_many)

    def clear_all(self, error_id: str):
        if error_id in self.error_state_map.keys():
            self.error_state_map[error_id].clear_all()

    def get_error_states(self) -> List[ErrorStateKeeper]:
        result: List[ErrorStateKeeper] = []
        for error_state in self.error_state_map.values():
            if error_state.is_set():
                result.append(error_state)
        return result


# --------------------------------------------------------------------------------
if __name__ == "__main__":
    print("Not an executable script, intended for use in other scripts")



import asyncio
import socket
import uuid

from typing import Type, Tuple
from pydantic import BaseModel as PydanticBaseModel

from ..persistent_stream_client_base import PersistentStreamClientBase

from ...data_transfer_objects import EmptyDto

# --------------------------------------------------------------------------------
class PersistedUDSClient(PersistentStreamClientBase):
    def __init__(
        self,
        server_path: str,
        max_retries: int   = 3,
        retry_delay: float = 0.1,
    ):
        self.server_path : str  = server_path
        self.can_transmit: bool = hasattr(socket, "AF_UNIX")
        super().__init__(max_retries, retry_delay)

    # --------------------------------------------------------------------------------
    async def open_connection(self) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        return await asyncio.open_unix_connection(self.server_path)

    # --------------------------------------------------------------------------------
    async def send_message(
        self,
        route_key        : str,
        data             : PydanticBaseModel,
        response_dto_type: Type[PydanticBaseModel] = EmptyDto,
        request_uid      : uuid.UUID               = None,
    ) -> PydanticBaseModel:
        if not self.can_transmit:
            raise Exception("UDS communications are not supported on this platform. Will not send message.")

        return await super().send_message(route_key, data, response_dto_type, request_uid)



import uuid
import re
import socket
import ipaddress

from typing import Dict, Any
from urllib.parse import urlparse

# --------------------------------------------------------------------------------
def camel_to_snake(input_string: str) -> str:
    output_string = ""
    for i in range(len(input_string)):
        if input_string[i].isupper() and i != 0:
            output_string += "_" + input_string[i].lower()
        else:
            output_string += input_string[i].lower()
    return output_string

# --------------------------------------------------------------------------------
def string_to_uuid(value: str) -> uuid.UUID | bool:
    try:
        return uuid.UUID(value)
    except ValueError:
        return False

# --------------------------------------------------------------------------------
def is_valid_ipv4_address(ip_address: str) -> bool:
    try:
        ipaddress.IPv4Address(ip_address)
        return True
    except ipaddress.AddressValueError:
        return False

# Checks hostname against Internet Engineering Task Force (IETF) standard
# --------------------------------------------------------------------------------
def is_possible_valid_ietf_hostname(hostname: str) -> bool:
    if len(hostname) > 255: # It may not be longer than 255 characters
        return False
    if hostname[-1] == ".":
        hostname = hostname[:-1]  # strip exactly one dot from the right, if present
    allowed = re.compile("(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)
    return all(allowed.match(x) for x in hostname.split("."))

# --------------------------------------------------------------------------------
def is_valid_hostname(hostname: str) -> bool:
    if not is_valid_ipv4_address(hostname):
        return is_possible_valid_ietf_hostname(hostname)
    return True

# --------------------------------------------------------------------------------
def is_valid_port_number(port_number: str) -> bool:
    return port_number.isdigit()

# --------------------------------------------------------------------------------
def flatten_dictionary(dictionary: Dict[str, Any], current_key: str = '') -> Dict[str, Any]:
    retval: Dict[str, Any] = {}
    for key, value in dictionary.items():
        new_key = f"{current_key}.{key}" if current_key else key
        if isinstance(dictionary[key], dict):
            retval.update(flatten_dictionary(value, new_key))
        else:
            retval[new_key] = value
    return retval

# --------------------------------------------------------------------------------
# This gets used to get the host name of a machine running an Ecosystem application
# --------------------------------------------------------------------------------
def get_machine_hostname(hostname: str) -> str:
    if hostname == "127.0.0.1" or hostname == "localhost":
        return socket.gethostname()
    elif is_valid_ipv4_address(hostname):
        return socket.gethostbyname(hostname)
    elif is_possible_valid_ietf_hostname(hostname):
        return hostname

# --------------------------------------------------------------------------------
def is_valid_url(url):
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False



import asyncio

from ..stream_server_base import StreamServerBase

from ...configuration.config_models import ConfigTCP


# --------------------------------------------------------------------------------
class TCPServer(StreamServerBase):
    def __init__(self, configuration : ConfigTCP):
        super().__init__()
        self.host: str = configuration.host
        self.port: int = configuration.port

    # --------------------------------------------------------------------------------
    async def __aenter__(self):
        try:
            self._running    = True
            self._logger.info(f"Setting up TCP server.")
            await self.__setup_server()
        except Exception as e:
            self._running = False
            self._server.close()
            raise e

    # --------------------------------------------------------------------------------
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    # --------------------------------------------------------------------------------
    def stop(self):
        if self._running:
            self._running = False
            self._logger.info(f"Stopping TCP server for {self.host}:{self.port}.")
            self._server.close()

    # --------------------------------------------------------------------------------
    async def __setup_server(self):
        self._server   = await asyncio.start_server(self._handle_request, self.host, self.port)
        serving_address = self._server.sockets[0].getsockname()
        self._logger.info(f'Serving TCP on {serving_address}')

    # --------------------------------------------------------------------------------
    async def serve(self):
        if self._running:
            async with self._server:
                try:
                    await self._server.serve_forever(),
                except asyncio.exceptions.CancelledError:
                    self._logger.info("TCP server cancelled.")


# --------------------------------------------------------------------------------
if __name__ == "__main__":
    print("Not an executable script, intended for use in other scripts")



import uuid
import json

from abc import ABC, abstractmethod
from typing import Type
from pydantic import BaseModel as PydanticBaseModel

from ..data_transfer_objects import RequestDTO, ResponseDTO, EmptyDto
from ..requests.status import Status

from ..exceptions import (
    ProtocolParsingException,
    ClientDeniedException,
    PydanticValidationException,
    RouteKeyUnknownException,
    ServerBusyException,
    ProcessingException,
    UnknownException,
    UnknownStatusCodeException,
)

# --------------------------------------------------------------------------------
class ClientBase(ABC):
    def __init__(
        self,
        max_retries: int   = 3,
        retry_delay: float = 0.1,
    ):
        self.max_retries: int   = max_retries
        self.retry_delay: float = retry_delay
        self.retry_count: int   = 0
        self.success    : bool  = False

    # --------------------------------------------------------------------------------
    @abstractmethod
    async def _send_message_retry_loop(self, request: str) -> str:
        pass

    # --------------------------------------------------------------------------------
    @staticmethod
    def generate_response_exception(request: ResponseDTO):
        if request.status == Status.PROTOCOL_PARSING_ERROR.value:
            return ProtocolParsingException(str(request.data))

        if request.status == Status.CLIENT_DENIED.value:
            return ClientDeniedException(str(request.data.json()))

        if request.status == Status.PYDANTIC_VALIDATION_ERROR.value:
            return PydanticValidationException(str(request.data))

        if request.status == Status.ROUTE_KEY_UNKNOWN.value:
            return RouteKeyUnknownException(str(request.data))

        if request.status == Status.APPLICATION_BUSY.value:
            return ServerBusyException(str(request.data))

        if request.status == Status.PROCESSING_FAILURE.value:
            return ProcessingException(str(request.data))

        if request.status == Status.UNKNOWN.value:
            return UnknownException(str(request.data))

        return UnknownStatusCodeException(f"Unrecognised status code[{request.status}]: {str(request.data)}")

    # --------------------------------------------------------------------------------
    async def send_message(
        self,
        route_key        : str,
        data             : PydanticBaseModel,
        response_dto_type: Type[PydanticBaseModel] = EmptyDto,
        request_uid      : uuid.UUID               = None
    ) -> PydanticBaseModel:
        if not request_uid:
            uuid_to_use = uuid.uuid4()
        else:
            uuid_to_use = request_uid

        self.success     = False
        self.retry_count = 0
        request          = RequestDTO(uid=str(uuid_to_use), route_key = route_key, data = data)
        request_str      = request.model_dump_json()
        response_str     = await self._send_message_retry_loop(f"{request_str}\n")
        response_dict    = json.loads(response_str)
        response         = ResponseDTO(**response_dict)

        if response.status != Status.SUCCESS.value:
            raise self.generate_response_exception(response)

        response_dto     = response_dto_type(**response.data)
        return response_dto



import json
import uuid
import sqlalchemy

from typing import Type, cast, TypeVar, Generic, List

from pydantic import BaseModel as PydanticBaseModel

from sqlalchemy import create_engine, Column, BigInteger, BINARY, String, func, Table, MetaData
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import sessionmaker


OrmBaseClass = sqlalchemy.orm.declarative_base()

_QueuedType = TypeVar('_QueuedType', bound=PydanticBaseModel)


# --------------------------------------------------------------------------------
class QueueRecord(OrmBaseClass):
    __tablename__ = 'queued_objects'
    record_id     = Column(BigInteger, primary_key=True)
    record_uuid   = Column(BINARY(16), unique=True, index=True, nullable=False, default=uuid.uuid4().bytes)
    object_string = Column(String)


# --------------------------------------------------------------------------------
class SqlPersistedQueueBase:
    def __init__(self, file_path: str):
        self.file_path        : str               = file_path
        self.connection_string: str               = f"sqlite:///{file_path}"
        self.database_engine  : sqlalchemy.Engine = create_engine(self.connection_string)

        OrmBaseClass.metadata.create_all(self.database_engine)

        self.session = sessionmaker(bind=self.database_engine, autoflush=False)()


# --------------------------------------------------------------------------------
class SqlPersistedQueue(Generic[_QueuedType], SqlPersistedQueueBase):
    def __init__(
        self,
        file_path     : str,
        queued_type   : Type[_QueuedType],
        max_uncommited: int = 0,
    ):
        super().__init__(file_path)
        self.max_uncommited  : int               = max_uncommited
        self.uncommited_count: int               = 0
        self.queued_type     : Type[_QueuedType] = queued_type

    # --------------------------------------------------------------------------------
    def shut_down(self):
        self.session.commit()
        self.uncommited_count = 0
        self.session.close()

    # --------------------------------------------------------------------------------
    # Note:
    # Because SQLite defines BIGINT as signed, and can deal with negative values
    # in a primary key. We genuinely don't have to be concerned about dealing with
    # a minimum record id here. Thankyou SQLite guys. I Love your work!
    async def push_front(self, object_to_queue: _QueuedType, uid: uuid.UUID = None) -> uuid.UUID :
        queued_record = self.__get_record_by_uuid(uid)
        if queued_record is not None:
            return uid

        queue_uuid = uid or uuid.uuid4()
        new_record = QueueRecord(
            record_uuid   = queue_uuid.bytes,
            object_string = object_to_queue.model_dump_json()
        )
        new_record.record_id = self.__get_min_record_id() - 1
        self.session.add(new_record)
        self.session.commit()
        return queue_uuid

    # --------------------------------------------------------------------------------
    async def push_back(self, object_to_queue: _QueuedType, uid: uuid.UUID = None) -> uuid.UUID:
        queued_record = self.__get_record_by_uuid(uid)
        if queued_record is not None:
            return uid

        queue_uuid = uid or uuid.uuid4()
        new_record = QueueRecord(
            record_uuid   = queue_uuid.bytes,
            object_string = object_to_queue.model_dump_json()
        )
        new_record.record_id = self.__get_max_record_id() + 1
        self.session.add(new_record)
        self.session.commit()
        return queue_uuid

    # --------------------------------------------------------------------------------
    async def pop_front(self):
        queued_record = self.__get_record_with_lowest_record_id()
        if queued_record is None:
            return None
        queueable_type_object = self.__record_to_queueable_object(queued_record)
        self.session.delete(queued_record)
        self.session.commit()
        return queueable_type_object

    # --------------------------------------------------------------------------------
    async def pop_back(self):
        queued_record = self.__get_record_with_highest_record_id()
        if queued_record is None:
            return None
        queueable_type_object = self.__record_to_queueable_object(queued_record)
        self.session.delete(queued_record)
        self.session.commit()
        return queueable_type_object

    # --------------------------------------------------------------------------------
    async def pop_uuid(self, object_uuid: uuid.UUID):
        queued_record = self.__get_record_by_uuid(object_uuid)
        if queued_record is None:
            return None
        queueable_type_object = self.__record_to_queueable_object(queued_record)
        self.session.delete(queued_record)
        self.session.commit()
        return queueable_type_object

    # --------------------------------------------------------------------------------
    async def inspect_front(self):
        queued_record = self.__get_record_with_lowest_record_id()
        if queued_record is None:
            return None
        return self.__record_to_queueable_object(queued_record)

    # --------------------------------------------------------------------------------
    async def inspect_back(self):
        queued_record = self.__get_record_with_highest_record_id()
        if queued_record is None:
            return None
        return self.__record_to_queueable_object(queued_record)

    # --------------------------------------------------------------------------------
    async def inspect_uuid(self, object_uuid: uuid.UUID):
        queued_record = self.__get_record_by_uuid(object_uuid)
        if queued_record is None:
            return None
        return self.__record_to_queueable_object(queued_record)

    # --------------------------------------------------------------------------------
    def size(self) -> int:
        # return int(self.session.query(func.count(QueueRecord.record_id)).scalar())
        return self.session.query(QueueRecord).count()

    # --------------------------------------------------------------------------------
    def is_empty(self) -> int:
        return self.size() == 0

    # --------------------------------------------------------------------------------
    async def get_first_x_uuids(self, how_many: int = 1) -> List[str]:
        response: List[str] = []
        queued_records      = self.session.query(QueueRecord).limit(how_many).all()
        for record in queued_records:
            response.append(str(uuid.UUID(bytes=record.record_uuid)))
        return response

    # --------------------------------------------------------------------------------
    async def clear(self):
        metadata = MetaData()
        table    = Table("queued_objects", metadata, autoload_with=self.database_engine)
        self.session.query(table).filter(table.c.record_id > 0).delete()
        self.session.query(table).filter(table.c.record_id < 0).delete()
        self.session.query(table).filter(table.c.record_id == 0).delete()
        self.session.commit()
        self.uncommited_count = 0

    # --------------------------------------------------------------------------------
    def __get_min_record_id(self) -> int:
        if self.size() < 1:
            return 0
        return int(self.session.query(func.min(QueueRecord.record_id)).scalar())

    # --------------------------------------------------------------------------------
    def __get_max_record_id(self) -> int:
        if self.size() < 1:
            return 0
        return int(self.session.query(func.max(QueueRecord.record_id)).scalar())

    # --------------------------------------------------------------------------------
    def __get_record_by_record_id(self, record_id: int) -> QueueRecord | None:
        try:
            # cast is used here to stop some IDEs from thinking that the query returns Type[QueueRecord] rather than QueueRecord
            return cast(
                QueueRecord,
                self.session.query(QueueRecord).filter(QueueRecord.record_id == record_id).one() # noqa PyCharm's report that the equality check has a problem is a false positive.
            )
        except NoResultFound:
            return None

    # --------------------------------------------------------------------------------
    def __get_record_by_uuid(self, record_uuid: uuid.UUID) -> QueueRecord | None:
        try:
            # cast is used here to stop some IDEs from thinking that the query returns Type[QueueRecord] rather than QueueRecord
            return cast(
                QueueRecord,
                self.session.query(QueueRecord).filter(QueueRecord.record_uuid == record_uuid.bytes).one() # noqa PyCharm's report that the equality check has a problem is a false positive.
            )
        except NoResultFound:
            return None

    # --------------------------------------------------------------------------------
    def __get_record_with_lowest_record_id(self) -> QueueRecord | None:
        if self.size() < 1:
            return None
        return self.__get_record_by_record_id(self.__get_min_record_id())

    # --------------------------------------------------------------------------------
    def __get_record_with_highest_record_id(self) -> QueueRecord | None:
        if self.size() < 1:
            return None
        return self.__get_record_by_record_id(self.__get_max_record_id())

    # --------------------------------------------------------------------------------
    def __record_to_queueable_object(self, record: QueueRecord) -> _QueuedType:
        queued_data = json.loads(record.object_string)
        return self.queued_type(**queued_data)

    # --------------------------------------------------------------------------------
    # async def __check_commit_count(self):
    #     # if self.uncommited_count >= self.max_uncommited:
    #     #     self.session.commit()
    #     #     self.uncommited_count = 0
    #     return



from typing import Type, TypeVar
from pydantic import BaseModel as PydanticBaseModel

from ..clients import ClientBase

_ResponseDTOType = TypeVar("_ResponseDTOType", bound=PydanticBaseModel)


def sender(
    client           : ClientBase,
    route_key        : str,
    response_dto_type: Type[_ResponseDTOType]
):
    def inner_decorator(function):
        async def wrapper(*args, **kwargs) -> _ResponseDTOType:
            request_uid_to_use = None
            if "request_uid" in kwargs.keys():
                request_uid_to_use = kwargs["request_uid"]

            request_dto  = await function(*args, **kwargs)
            response_dto = await client.send_message(route_key, request_dto, response_dto_type, request_uid_to_use)
            return response_dto
        return wrapper
    return inner_decorator



import json

from ..util import SingletonType


# --------------------------------------------------------------------------------
class ErrorStateKeeper(metaclass=SingletonType):
    def __init__(self, error_id: str, description: str):
        self._error_id   : str = error_id
        self._description: str = description
        self._error_count: int = 0

    def to_dict(self):
        return {
            'error_id'   : self._error_id,
            'description': self._description,
            'count'      : self._error_count
        }

    def to_json(self):
        return json.dumps(self.to_dict())

    def __str__(self):
        return f"ErrorState: {self.to_json()}"

    def increment(self):
        self._error_count += 1

    def clear_some(self, how_many: int):
        if self._error_count >= how_many:
            self._error_count = self._error_count - how_many
        else:
            self._error_count = 0

    def clear_all(self):
        self._error_count = 0

    def get_error_count(self):
        return self._error_count

    def is_set(self):
        return self._error_count > 0

    def get_id(self):
        return self._error_id

    def get_description(self):
        return self._description


# --------------------------------------------------------------------------------
if __name__ == "__main__":
    print("Not an executable script, intended for use in other scripts")



import asyncio

FIRE_AND_FORGET_TASKS = set()

# --------------------------------------------------------------------------------
def fire_and_forget_task_for_loop(loop: asyncio.AbstractEventLoop, coroutine):
    global FIRE_AND_FORGET_TASKS
    task = loop.create_task(coroutine)
    FIRE_AND_FORGET_TASKS.add(task)
    task.add_done_callback(FIRE_AND_FORGET_TASKS.discard)
    return task

# --------------------------------------------------------------------------------
def fire_and_forget_task(coroutine):
    global FIRE_AND_FORGET_TASKS
    task = asyncio.create_task(coroutine)
    FIRE_AND_FORGET_TASKS.add(task)
    # Should this rather be:
    # task.add_done_callback(lambda x: FIRE_AND_FORGET_TASKS.discard(x))
    task.add_done_callback(FIRE_AND_FORGET_TASKS.discard)
    return task

# --------------------------------------------------------------------------------
def run_soon(function):
    def called_task(loop: asyncio.AbstractEventLoop, *args, **kwargs):
        task = loop.create_task(function(*args, **kwargs))
        FIRE_AND_FORGET_TASKS.add(task)
        task.add_done_callback(FIRE_AND_FORGET_TASKS.discard)
        return task

    def wrapper(*args, **kwargs):
        loop = asyncio.get_running_loop()
        loop.call_soon(called_task, loop, *args, **kwargs)
    return wrapper



from .json import TCPClient
from .json import UDPClient
from .json import UDSClient
from .json.persisted_tcp import PersistedTCPClient
from .json.persisted_uds import PersistedUDSClient
from .client_base import ClientBase



import uuid

from typing import Dict, List, Any

from ..sending.queued_sender_class import QueuedSenderClass
from ..util import SingletonType


# --------------------------------------------------------------------------------
class QueuedSenderKeeper(metaclass=SingletonType):
    __queued_senders: Dict[str, QueuedSenderClass] = {}

    def __init__(self):
        pass

    # --------------------------------------------------------------------------------
    def get_queued_senders(self) -> List[QueuedSenderClass]:
        return [x for x in self.__queued_senders.values()]

    # --------------------------------------------------------------------------------
    def has_route_key(self, route_key: str) -> bool:
        if route_key not in self.__queued_senders.keys():
            return False
        return True

    # --------------------------------------------------------------------------------
    def add_queued_sender(self, queued_sender: QueuedSenderClass):
        self.__queued_senders[queued_sender.get_route_key()] = queued_sender

    # --------------------------------------------------------------------------------
    async def get_queue_information(self, route_key: str) -> Dict[str, Any] | None:
        if route_key not in self.__queued_senders.keys():
            return None
        return await self.__get_queue_information(route_key)

    # --------------------------------------------------------------------------------
    async def __get_queue_information(self, route_key: str) -> Dict[str, Any]:
        return {
            "pending"             : await self.__queued_senders[route_key].pending_queue_size(),
            "error"               : await self.__queued_senders[route_key].error_queue_size(),
            "send_process_paused" : self.__queued_senders[route_key].is_send_process_paused(),
        }

    # --------------------------------------------------------------------------------
    async def pause_send_process(self, route_key: str):
        if route_key not in self.__queued_senders.keys():
            return None
        self.__queued_senders[route_key].pause_send_process()
        return await self.__get_queue_information(route_key)

    # --------------------------------------------------------------------------------
    async def un_pause_send_process(self, route_key: str):
        if route_key not in self.__queued_senders.keys():
            return None
        self.__queued_senders[route_key].unpause_send_process()
        return await self.__get_queue_information(route_key)

    # Request popping
    # --------------------------------------------------------------------------------
    async def pop_request_from_error_queue(self, route_key: str, request_uid: uuid.UUID):
        if route_key not in self.__queued_senders.keys():
            return None
        request = await self.__queued_senders[route_key].pop_request_from_error_queue(request_uid)
        if request is None:
            return False
        queue_information = await self.__get_queue_information(route_key)
        queue_information["request_data"] = request
        return queue_information

    # Request inspection
    # --------------------------------------------------------------------------------
    async def inspect_request_in_error_queue(self, route_key: str, request_uid: uuid.UUID,):
        if route_key not in self.__queued_senders.keys():
            return None
        request = await self.__queued_senders[route_key].inspect_request_in_error_queue(request_uid)
        if request is None:
            return False
        queue_information = await self.__get_queue_information(route_key)
        queue_information["request_data"] = request
        return queue_information

    # --------------------------------------------------------------------------------
    async def get_first_x_error_uuids(self, route_key) -> List[str] | None:
        if route_key not in self.__queued_senders.keys():
            return None
        return await self.__queued_senders[route_key].get_first_x_error_uuids(10)

    # --------------------------------------------------------------------------------
    async def reprocess_error_queue(self, route_key: str):
        if route_key not in self.__queued_senders.keys():
            return None
        await self.__queued_senders[route_key].reprocess_error_queue()
        return await self.__get_queue_information(route_key)

    # --------------------------------------------------------------------------------
    async def clear_error_queue(self, route_key: str):
        if route_key not in self.__queued_senders.keys():
            return None
        await self.__queued_senders[route_key].clear_queue_database('error')
        return await self.__get_queue_information(route_key)

    # --------------------------------------------------------------------------------
    async def reprocess_error_queue_request_uid(self, route_key: str, request_uid: uuid.UUID):
        if route_key not in self.__queued_senders.keys():
            return None
        result = await self.__queued_senders[route_key].reprocess_error_queue_request_uid(request_uid)
        if not result:
            return False
        queue_information = await self.__get_queue_information(route_key)
        queue_information["request_data"] = result.dict()
        return queue_information



import json
import uuid
import sqlalchemy

from typing import cast, Dict, List, Type, TypeVar, Generic

from pydantic import BaseModel as PydanticBaseModel

from sqlalchemy import create_engine, Column, BigInteger, BINARY, String, func, Table, MetaData, asc, desc, exists
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import sessionmaker

OrmBaseClass = sqlalchemy.orm.declarative_base()

_QueuedType = TypeVar('_QueuedType', bound=PydanticBaseModel)

# --------------------------------------------------------------------------------
class QueueRecord(OrmBaseClass):
    __tablename__ = 'queued_objects'
    record_id     = Column(BigInteger, primary_key=True)
    record_uuid   = Column(BINARY(16), unique=True, index=True, nullable=False, default=uuid.uuid4().bytes)
    object_string = Column(String)

# --------------------------------------------------------------------------------
class PageEntry(PydanticBaseModel):
    record_uuid  : uuid.UUID
    object_string: str

# --------------------------------------------------------------------------------
class QueuePage:
    def __init__(self, queued_type: Type[_QueuedType]):
        self.__page_data_dict: Dict[uuid.UUID, PageEntry] = {}
        self.__page_data_list: List[PageEntry]            = []
        self.__queued_type   : Type[_QueuedType]          = queued_type

    def __prepend_entry(self, entry: PageEntry):
        self.__page_data_list.insert(0, entry)
        self.__page_data_dict[entry.record_uuid] = entry

    def __append_entry(self, entry: PageEntry):
        self.__page_data_list.append(entry)
        self.__page_data_dict[entry.record_uuid] = entry

    @staticmethod
    def __make_entry(object_to_queue: _QueuedType, uid: uuid.UUID = None) -> PageEntry:
        queue_uuid = uid or uuid.uuid4()
        return PageEntry(
            record_uuid   = queue_uuid.bytes,
            object_string = object_to_queue.model_dump_json()
        )

    def __entry_to_queueable_object(self, entry: PageEntry) -> _QueuedType:
        queued_data = json.loads(entry.object_string)
        return self.__queued_type(**queued_data)

    def push(self, object_to_queue: _QueuedType, uid: uuid.UUID = None) -> uuid.UUID:
        return self.push_back(object_to_queue, uid)

    def push_with_str(self, data: str, uid: uuid.UUID) -> uuid.UUID:
        entry = PageEntry(
            record_uuid   = uid,
            object_string = data
        )
        self.__append_entry(entry)
        return uid

    def pop(self) -> _QueuedType:
        return self.pop_front()

    def push_back(self, object_to_queue: _QueuedType, uid: uuid.UUID = None) -> uuid.UUID:
        entry = self.__make_entry(object_to_queue, uid)
        self.__append_entry(entry)
        return entry.record_uuid

    def push_front(self, object_to_queue: _QueuedType, uid: uuid.UUID = None):
        entry = self.__make_entry(object_to_queue, uid)
        self.__prepend_entry(entry)
        return entry.record_uuid

    def pop_front(self):
        if self.size() < 1:
            return None
        entry = self.__page_data_list.pop(0)
        self.__page_data_dict.pop(entry.record_uuid)
        return self.__entry_to_queueable_object(entry)

    def pop_back(self):
        if self.size() < 1:
            return None
        entry = self.__page_data_list.pop()
        self.__page_data_dict.pop(entry.record_uuid)
        return self.__entry_to_queueable_object(entry)

    def size(self):
        return len(self.__page_data_list)

    def get_page_list(self):
        return self.__page_data_list

    def has_entry(self, uid: uuid.UUID):
        return uid in self.__page_data_dict.keys()

    def inspect_entry(self, uid: uuid.UUID):
        if uid not in self.__page_data_dict.keys():
            return None

        entry = self.__page_data_dict[uid]
        return self.__entry_to_queueable_object(entry)

    def pop_entry(self, uid: uuid.UUID):
        if uid not in self.__page_data_dict.keys():
            return None

        entry           = self.__page_data_dict[uid]
        index_to_remove = -1
        for i in range(len(self.__page_data_list)):
            if entry.record_uuid == self.__page_data_list[i].record_uuid:
                index_to_remove = i
        self.__page_data_list.pop(index_to_remove)
        return entry

    def get_first_x_uuids(self, how_many: int):
        retval: List[str] = []
        for x in range(how_many):
            retval.append(str(self.__page_data_list[x].record_uuid))
        return retval

# --------------------------------------------------------------------------------
class PaginatedQueue(Generic[_QueuedType]):
    def __init__(
        self,
        file_path  : str,
        queued_type: Type[_QueuedType],
        page_size  : int = 100,
    ):
        self.file_path        : str               = file_path
        self.connection_string: str               = f"sqlite:///{file_path}"
        self.database_engine  : sqlalchemy.Engine = create_engine(self.connection_string)
        self.queued_type      : Type[_QueuedType] = queued_type
        self.page_size        : int               = page_size
        self.front_page       : QueuePage         = QueuePage(queued_type)
        self.back_page        : QueuePage         = self.front_page

        OrmBaseClass.metadata.create_all(self.database_engine)

        self.session = sessionmaker(bind=self.database_engine, autoflush=False)()
        self.__do_initial_load()

    # --------------------------------------------------------------------------------
    def __do_initial_load(self):
        if self.__get_database_size() > 0:
            self.__load_front_page()

        if self.__get_database_size() > 0:
            self.back_page = QueuePage(self.queued_type)
            self.__load_back_page()

    # --------------------------------------------------------------------------------
    def __get_max_record_id(self) -> int:
        if self.__get_database_size() < 1:
            return 0
        return int(self.session.query(func.max(QueueRecord.record_id)).scalar())

    # --------------------------------------------------------------------------------
    def __get_min_record_id(self) -> int:
        if self.__get_database_size() < 1:
            return 0
        return int(self.session.query(func.min(QueueRecord.record_id)).scalar())

    # --------------------------------------------------------------------------------
    def __delete_record_ids(self, record_ids: List[int]):
        self.session.\
            query(QueueRecord).\
            filter(QueueRecord.record_id.in_(record_ids)).\
            delete(synchronize_session='fetch')
        self.session.commit()

    # --------------------------------------------------------------------------------
    def __load_front_page(self):
        records = self.session.\
            query(QueueRecord).\
            order_by(asc(QueueRecord.record_id)).\
            limit(self.page_size).all()

        for record in records:
            self.front_page.push_with_str(record.object_string, uuid.UUID(bytes=record.record_uuid))
        ids_to_delete = [x.record_id for x in records]
        self.__delete_record_ids(ids_to_delete)

    # --------------------------------------------------------------------------------
    def __load_back_page(self):
        records = self.session.\
            query(QueueRecord).\
            order_by(desc(QueueRecord.record_id)).\
            limit(self.page_size).all()

        for record in records:
            self.back_page.push_with_str(record.object_string, uuid.UUID(bytes=record.record_uuid))
        ids_to_delete = [x.record_id for x in records]
        self.__delete_record_ids(ids_to_delete)

    # --------------------------------------------------------------------------------
    def __write_front_page(self):
        page_data     = list(reversed(self.front_page.get_page_list()))
        min_record_id = self.__get_min_record_id() - 1
        data_to_write = []
        for entry in page_data:
            data_to_write.append(QueueRecord(
                record_id    = min_record_id,
                record_uuid  = entry.record_uuid.bytes,
                object_string= entry.object_string
            ))
            min_record_id -= 1
        self.session.bulk_save_objects(data_to_write)
        self.session.commit()
        self.front_page = QueuePage(self.queued_type)

    # --------------------------------------------------------------------------------
    def __write_back_page(self):
        page_data     = self.back_page.get_page_list()
        max_record_id = self.__get_max_record_id() + 1
        data_to_write = []
        for entry in page_data:
            data_to_write.append(QueueRecord(
                record_id    = max_record_id,
                record_uuid  = entry.record_uuid.bytes,
                object_string= entry.object_string
            ))
            max_record_id += 1
        self.session.bulk_save_objects(data_to_write)
        self.session.commit()
        self.back_page = QueuePage(self.queued_type)

    # --------------------------------------------------------------------------------
    def __get_database_size(self):
        return self.session.query(QueueRecord).count()

    # --------------------------------------------------------------------------------
    def __get_record_by_uuid(self, uid: uuid.UUID) -> QueueRecord | None:
        try:
            record = self.session.\
                query(QueueRecord).\
                filter(QueueRecord.record_uuid == uid.bytes).\
                one()
            return cast(QueueRecord, record)
        except NoResultFound:
            return None

    # --------------------------------------------------------------------------------
    def __check_record_exists(self, uid: uuid.UUID) -> bool:
        return self.session.\
            query(exists().where(QueueRecord.record_uuid == uid.bytes)).\
            scalar()

    # --------------------------------------------------------------------------------
    async def pop(self):
        if self.front_page.size() > 0:
            return self.front_page.pop()
        elif self.__get_database_size() > 0:
            self.__load_front_page()
            return self.front_page.pop()
        elif self.back_page.size() > 0:
            self.front_page = self.back_page
            return self.front_page.pop()
        else:
            self.front_page = self.back_page
            return None

    # --------------------------------------------------------------------------------
    async def push(self, object_to_queue: _QueuedType, uid: uuid.UUID = None) -> uuid.UUID:
        if (self.back_page.has_entry(uid)  or
            self.front_page.has_entry(uid) or
            self.__check_record_exists(uid)):
            return uid

        if self.back_page.size() < self.page_size:
            return self.back_page.push(object_to_queue, uid)
        elif self.back_page is self.front_page:
            self.back_page = QueuePage(self.queued_type)
            return self.back_page.push(object_to_queue, uid)
        else:
            self.__write_back_page()
            return self.back_page.push(object_to_queue, uid)

    # --------------------------------------------------------------------------------
    def size(self):
        total = self.__get_database_size()
        if self.front_page is self.back_page:
            total += self.front_page.size()
        else:
            front_size = self.front_page.size()
            back_size  = self.back_page.size()
            total     += front_size + back_size
        return total

    # --------------------------------------------------------------------------------
    async def inspect_uuid(self, uid: uuid.UUID):
        if self.front_page.has_entry(uid):
            return self.front_page.inspect_entry(uid)

        if self.back_page.has_entry(uid):
            return self.back_page.inspect_entry(uid)

        record = self.__get_record_by_uuid(uid)
        if record is None:
            return None
        queued_data = json.loads(record.object_string)
        return self.queued_type(**queued_data)

    # --------------------------------------------------------------------------------
    async def pop_uuid(self, uid: uuid.UUID):
        if self.front_page.has_entry(uid):
            return self.front_page.pop_entry(uid)

        if self.back_page.has_entry(uid):
            return self.back_page.pop_entry(uid)

        record = self.__get_record_by_uuid(uid)
        if record is None:
            return None
        queued_data = json.loads(record.object_string)
        self.session.delete(record)
        self.session.commit()
        return self.queued_type(**queued_data)

    # --------------------------------------------------------------------------------
    def is_empty(self) -> int:
        return self.size() == 0

    # --------------------------------------------------------------------------------
    async def get_first_x_uuids(self, how_many: int = 1) -> List[str]:
        return self.front_page.get_first_x_uuids(how_many)

    # --------------------------------------------------------------------------------
    async def clear(self):
        metadata = MetaData()
        table    = Table("queued_objects", metadata, autoload_with=self.database_engine)
        self.session.query(table).filter(table.c.record_id > 0).delete()
        self.session.query(table).filter(table.c.record_id < 0).delete()
        self.session.query(table).filter(table.c.record_id == 0).delete()
        self.session.commit()
        self.front_page = QueuePage(self.queued_type)
        self.back_page  = self.front_page

    # --------------------------------------------------------------------------------
    def shut_down(self):
        if self.front_page.size() > 0:
            self.__write_front_page()

        if self.back_page.size() > 0:
            self.__write_back_page()

        self.session.close()



import uuid

from typing import Type, TypeVar, Generic
from pydantic import BaseModel as PydanticBaseModel

from .queued_sender_base import QueuedSenderBase

from ..clients import ClientBase
from ..data_transfer_objects import EmptyDto

_RequestDTOType  = TypeVar("_RequestDTOType" , bound=PydanticBaseModel)
_ResponseDTOType = TypeVar("_ResponseDTOType", bound=PydanticBaseModel)


class QueuedSenderClass(Generic[_RequestDTOType, _ResponseDTOType], QueuedSenderBase[_RequestDTOType, _ResponseDTOType]):
    def __init__(
        self,
        client           : ClientBase,
        route_key        : str,
        request_dto_type : Type[_RequestDTOType],
        response_dto_type: Type[_ResponseDTOType] = EmptyDto,
        wait_period      : float                  = 0,
        page_size        : int                    = 100,
        max_retries      : int                    = 0,
    ):
        super().__init__(
            client,
            route_key,
            request_dto_type,
            response_dto_type,
            wait_period,
            page_size,
            max_retries
        )

    async def push_message(self, message: _RequestDTOType, request_uuid: uuid.UUID = None):
        await self.enqueue(message, request_uuid)



from ..datagram_client_base import DatagramClientBase

# --------------------------------------------------------------------------------
class UDPClient(DatagramClientBase):
    def __init__(
        self,
        server_host: str,
        server_port: int,
        timeout    : int   = 5,
        max_retries: int   = 3,
        retry_delay: float = 0.1,
    ):
        super().__init__(server_host, server_port, timeout, max_retries, retry_delay)

# --------------------------------------------------------------------------------
if __name__ == "__main__":
    print("Not an executable script, intended for use in other scripts")



import logging
import sys

from .buffered_rotating_file_handler import BufferedRotatingFileHandler

from ..configuration.config_models import AppConfiguration, ConfigLogging
from ..util import SingletonType

# --------------------------------------------------------------------------------
class EcoLogger(metaclass=SingletonType):
    __app_config        : AppConfiguration            = AppConfiguration()
    __log_config        : ConfigLogging               = __app_config.logging
    __level             : int                         = logging.DEBUG
    __logger            : logging.Logger              = None
    __formatter         : logging.Formatter           = None
    __file_handler      : BufferedRotatingFileHandler = None
    __console_handler                                 = None

    # --------------------------------------------------------------------------------
    def __setup_file_logging(self):
        log_file_config = self.__log_config.file_logging
        self.__file_handler = BufferedRotatingFileHandler(
            log_file_config.base_file_path,
            buffer_size  = log_file_config.buffer_size,
            max_bytes    = log_file_config.max_size_in_bytes,
            backup_count = log_file_config.max_files
        )

        self.__file_handler.setLevel(self.__level)
        self.__file_handler.setFormatter(self.__formatter)
        self.__logger.addHandler(self.__file_handler)

    # --------------------------------------------------------------------------------
    def __setup_console_logging(self):
        self.__console_handler = logging.StreamHandler(sys.stdout)
        self.__console_handler.setLevel(self.__level)
        self.__console_handler.setFormatter(self.__formatter)
        self.__logger.addHandler(self.__console_handler)

    # --------------------------------------------------------------------------------
    def __set_level(self, level: int):
        self.__level = level

        if self.__file_handler:
            self.__file_handler.setLevel(level)

        if self.__console_handler:
            self.__console_handler.setLevel(level)

        self.__logger.setLevel(level)

    # --------------------------------------------------------------------------------
    def set_level(self, level: str):
        if level == 'debug':
            self.__set_level(logging.DEBUG)

        if level == 'info':
            self.__set_level(logging.INFO)

        if level == 'warn':
            self.__set_level(logging.WARNING)

        if level == 'error':
            self.__set_level(logging.ERROR)

        if level == 'critical':
            self.__set_level(logging.CRITICAL)

    # --------------------------------------------------------------------------------
    def setup(self):
        log_file_config                = self.__log_config.file_logging
        log_file_config.base_file_name = f"{self.__app_config.name}-{self.__app_config.instance}"
        log_file_config.base_file_path = f"{log_file_config.directory}/{log_file_config.base_file_name}.log"
        self.__logger                  = logging.getLogger() # Yes, we are configuring the root logger!
        self.__formatter               = logging.Formatter(self.__log_config.format, datefmt=self.__log_config.date_format)

        if not self.__log_config.console_only:
            self.__setup_file_logging()

        if not self.__log_config.file_only:
            self.__setup_console_logging()

        self.__logger.setLevel(self.__level)
        # This switches asyncio logging to level WARNING
        # logging.getLogger('asyncio').setLevel(logging.WARNING)

    # --------------------------------------------------------------------------------
    def flush(self):
        if not self.__log_config.file_only:
            self.__console_handler.flush()

        if not self.__log_config.console_only:
            self.__file_handler.flush()

# --------------------------------------------------------------------------------
ekosis_logger = EcoLogger()
ekosis_logger.setup()



# --------------------------------------------------------------------------------
class ExceptionBase(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{type(self).__name__}: {self.message}"



import asyncio
from typing import Tuple, List

from .argument_parser import command_line_args, argument_parser
from .senders.eco_statistics_get import EcoStatisticsGetSender
from .senders.queued_endpoint_management import (
    EcoQueuedHandlerDataSender,
    EcoQueuedHandlerReceivingPauseSender,
    EcoQueuedHandlerProcessingPauseSender,
    EcoQueuedHandlerAllPauseSender,
    EcoQueuedHandlerReceivingUnPauseSender,
    EcoQueuedHandlerProcessingUnPauseSender,
    EcoQueuedHandlerAllUnPauseSender,
    EcoQueuedHandlerErrorsGetFirst10Sender,
    EcoQueuedHandlerErrorsReprocessAllSender,
    EcoQueuedHandlerErrorsClearSender,
    EcoQueuedHandlerErrorsReprocessOneSender,
    EcoQueuedHandlerErrorsPopRequestSender,
    EcoQueuedHandlerErrorsInspectRequestSender
)
from .senders.queued_sender_management import (
    EcoQueuedSenderDataSender,
    EcoQueuedSenderSendProcessPauseSender,
    EcoQueuedSenderSendProcessUnPauseSender,
    EcoQueuedSenderErrorsGetFirst10Sender,
    EcoQueuedSenderErrorsReprocessAllSender,
    EcoQueuedSenderErrorsClearSender,
    EcoQueuedSenderErrorsReprocessOneSender,
    EcoQueuedSenderErrorsPopRequestSender,
    EcoQueuedSenderErrorsInspectRequestSender,
)


from ..clients import TCPClient, UDPClient, UDSClient, ClientBase


# --------------------------------------------------------------------------------
def make_tcp_udp_host_port(request: str) -> Tuple[str, int]:
    host, port = request.split(":")
    if not host:
        print(f"Could not extract host from command: {request}")

    if not port:
        print(f"Could not extract port from command: {request}")

    return host, int(port)


# --------------------------------------------------------------------------------
def get_client() -> ClientBase:
    if command_line_args.server_type == "tcp":
        host, port = make_tcp_udp_host_port(command_line_args.server_details)
        client     = TCPClient(host, port)
    elif command_line_args.server_type == "udp":
        host, port = make_tcp_udp_host_port(command_line_args.server_details)
        client     = UDPClient(host, port)
    else:
        client = UDSClient(command_line_args.server_details)

    return client


# --------------------------------------------------------------------------------
def print_error(message: str):
    print("================================================================================")
    print(f"ERROR: {message}")
    print("--------------------------------------------------------------------------------")
    argument_parser.print_help()
    print("================================================================================")


# --------------------------------------------------------------------------------
def any_true_check(list_to_check: List[bool]) -> bool:
    for list_item in list_to_check:
        if list_item:
            return True
    return False


# --------------------------------------------------------------------------------
async def do_endpoint_action(client: ClientBase, route_key: str):
    if command_line_args.data or not any_true_check([
        command_line_args.data,
        command_line_args.pause_receiving,
        command_line_args.pause_processing,
        command_line_args.pause_all,
        command_line_args.unpause_receiving,
        command_line_args.unpause_processing,
        command_line_args.unpause_all,
        command_line_args.error_10,
        command_line_args.reprocess_all,
        command_line_args.clear,
        command_line_args.reprocess_one,
        command_line_args.inspect_request,
        command_line_args.pop_request,
    ]):
        await EcoQueuedHandlerDataSender(client).send(route_key)
        return

    if command_line_args.pause_receiving:
        await EcoQueuedHandlerReceivingPauseSender(client).send(route_key)
        return
    if command_line_args.pause_processing:
        await EcoQueuedHandlerProcessingPauseSender(client).send(route_key)
        return
    if command_line_args.pause_all:
        await EcoQueuedHandlerAllPauseSender(client).send(route_key)
        return
    if command_line_args.unpause_receiving:
        await EcoQueuedHandlerReceivingUnPauseSender(client).send(route_key)
        return
    if command_line_args.unpause_processing:
        await EcoQueuedHandlerProcessingUnPauseSender(client).send(route_key)
        return
    if command_line_args.unpause_all:
        await EcoQueuedHandlerAllUnPauseSender(client).send(route_key)
        return
    if command_line_args.error_10:
        await EcoQueuedHandlerErrorsGetFirst10Sender(client).send(route_key)
        return
    if command_line_args.reprocess_all:
        await EcoQueuedHandlerErrorsReprocessAllSender(client).send(route_key)
        return
    if command_line_args.clear:
        await EcoQueuedHandlerErrorsClearSender(client).send(route_key)
        return

    if not command_line_args.request_uid:
        print_error("--request_uid required!")
        return

    request_uid = command_line_args.request_uid

    if command_line_args.reprocess_one:
        await EcoQueuedHandlerErrorsReprocessOneSender(client).send(route_key, request_uid)
        return
    if command_line_args.inspect_request:
        await EcoQueuedHandlerErrorsInspectRequestSender(client).send(route_key, request_uid)
        return
    if command_line_args.pop_request:
        await EcoQueuedHandlerErrorsPopRequestSender(client).send(route_key, request_uid)
        return


# --------------------------------------------------------------------------------
async def do_sender_action(client: ClientBase, route_key: str):
    if (command_line_args.pause_receiving   or
        command_line_args.pause_processing  or
        command_line_args.unpause_receiving or
        command_line_args.unpause_processing):
        print_error("Pause and UN-Pause for receiving and processing, is only for queued endpoints!")

    if command_line_args.data or not any_true_check([
        command_line_args.data,
        command_line_args.pause_all,
        command_line_args.unpause_all,
        command_line_args.error_10,
        command_line_args.reprocess_all,
        command_line_args.clear,
        command_line_args.reprocess_one,
        command_line_args.inspect_request,
        command_line_args.pop_request,
    ]):
        await EcoQueuedSenderDataSender(client).send(route_key)
        return

    if command_line_args.pause_all:
        await EcoQueuedSenderSendProcessPauseSender(client).send(route_key)
        return
    if command_line_args.unpause_all:
        await EcoQueuedSenderSendProcessUnPauseSender(client).send(route_key)
        return
    if command_line_args.error_10:
        await EcoQueuedSenderErrorsGetFirst10Sender(client).send(route_key)
        return
    if command_line_args.reprocess_all:
        await EcoQueuedSenderErrorsReprocessAllSender(client).send(route_key)
        return
    if command_line_args.clear:
        await EcoQueuedSenderErrorsClearSender(client).send(route_key)
        return

    if not command_line_args.request_uid:
        print_error("--request_uid required!")
        return

    request_uid = command_line_args.request_uid

    if command_line_args.reprocess_one:
        await EcoQueuedSenderErrorsReprocessOneSender(client).send(route_key, request_uid)
        return
    if command_line_args.inspect_request:
        await EcoQueuedSenderErrorsInspectRequestSender(client).send(route_key, request_uid)
        return
    if command_line_args.pop_request:
        await EcoQueuedSenderErrorsPopRequestSender(client).send(route_key, request_uid)
        return


# --------------------------------------------------------------------------------
async def do_queue_action(client: ClientBase):
    if not command_line_args.route_key:
        print_error("--route_key required!")
        return

    route_key = command_line_args.route_key

    if command_line_args.action == "qem":
        await do_endpoint_action(client, route_key)
    else:
        await do_sender_action(client, route_key)


# --------------------------------------------------------------------------------
async def main():
    client = get_client()
    if command_line_args.action != "stat":
        await do_queue_action(client)
    else:
        await EcoStatisticsGetSender(client).send(command_line_args.statistics_type)

asyncio.run(main())
